from __future__ import annotations

from typing import NamedTuple, Tuple

import numba as nb
import numpy as np
import pandas as pd
from scipy import sparse

from xugrid.constants import (
    FILL_VALUE,
    BoolArray,
    FloatArray,
    IntArray,
    IntDType,
    SparseMatrix,
)


def cross2d(arr1: np.ndarray, arr2: np.ndarray) -> np.ndarray:
    """Cross product for arrays with 2D coordinates."""
    return arr1[..., 0] * arr2[..., 1] - arr1[..., 1] * arr2[..., 0]


def argsort_rows(array: np.ndarray) -> IntArray:
    if array.ndim != 2:
        raise ValueError(f"Array is not 2D, but has shape: {array.shape}")
    dtype = [("f{i}".format(i=i), array.dtype) for i in range(array.shape[1])]
    arr1d = array.view(dtype).flatten()
    return np.argsort(arr1d)


def index_like(xy_a: FloatArray, xy_b: FloatArray, tolerance: float):
    """Return the index that would transform xy_a into xy_b."""
    if xy_a.shape != xy_b.shape:
        raise ValueError("coordinates do not match in shape")

    # Quantize coordinates before sorting in case a tolerance is given.
    # After quantization, coordinates can be considered integers.
    # Float64 has ample room for our coordinate system even with small
    # tolerances, e.g. 1.0e9 / 1.0e-9 is not a problem.
    if tolerance != 0.0:
        sorter_a = argsort_rows(np.round(xy_a / tolerance))
        sorter_b = argsort_rows(np.round(xy_b / tolerance))
    else:
        sorter_a = argsort_rows(xy_a)
        sorter_b = argsort_rows(xy_b)

    if not np.allclose(xy_a[sorter_a], xy_b[sorter_b], rtol=0.0, atol=tolerance):
        raise ValueError("coordinates are not identical after sorting")
    #
    #   sorter_a         inverse_b
    # a --------> sorted ---------> b
    #
    inverse_b = np.argsort(sorter_b)
    return sorter_a[inverse_b]


class AdjacencyMatrix(NamedTuple):
    indices: IntArray
    indptr: IntArray
    nnz: int
    n: int
    m: int


def _csr_to_adjacency(A: sparse.csr_matrix) -> AdjacencyMatrix:
    if not isinstance(A, sparse.csr_matrix):
        raise TypeError(
            f"Expected scipy.sparse.csr_matrix, received: {type(A).__name__}"
        )

    n, m = A.shape
    adj = AdjacencyMatrix(A.indices, A.indptr, A.nnz, n, m)
    return adj


@nb.njit(inline="always")
def neighbors(A: AdjacencyMatrix, cell: int) -> IntArray:
    start = A.indptr[cell]
    end = A.indptr[cell + 1]
    return A.indices[start:end]


@nb.njit(inline="always")
def pop(array, size):
    return array[size - 1], size - 1


@nb.njit(inline="always")
def push(array, value, size):
    array[size] = value
    return size + 1


@nb.njit
def _topological_sort_by_dfs(A: AdjacencyMatrix):
    # This code is almost a direct port of the BSD-2 licensed code in Graphs.jl:
    #
    # https://github.com/JuliaGraphs/Graphs.jl/blob/master/LICENSE.md
    #
    # Copyright (c) 2015: Seth Bromberger and other contributors. Copyright (c)
    # 2012: John Myles White and other contributors.
    #
    # Redistribution and use in source and binary forms, with or without
    # modification, are permitted provided that the following conditions are met:
    #
    # Redistributions of source code must retain the above copyright notice, this
    # list of conditions and the following disclaimer. Redistributions in binary
    # form must reproduce the above copyright notice, this list of conditions and
    # the following disclaimer in the documentation and/or other materials provided
    # with the distribution. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    # CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
    # NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    # PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    # CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    # EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    # PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
    # OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
    # WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
    # OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    # ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    vcolor = np.zeros(A.m, dtype=np.uint8)
    verts = np.empty(A.m, dtype=np.int64)
    verts_size = 0
    S = np.empty(A.m, dtype=np.int64)
    S_size = 0

    for v in range(A.m):
        if vcolor[v] != 0:
            continue
        S_size = 0
        S_size = push(S, v, S_size)
        vcolor[v] = 1
        while S_size > 0:
            u = S[S_size - 1]
            w = 0
            for n in neighbors(A, u):
                if vcolor[n] == 1:
                    raise ValueError("The graph contains at least one cycle")
                elif vcolor[n] == 0:
                    w = n
                    break

            if w != 0:
                vcolor[w] = 1
                S_size = push(S, w, S_size)
            else:
                vcolor[u] = 2
                verts_size = push(verts, u, verts_size)
                S_size -= 1

    return verts[::-1]


def topological_sort_by_dfs(A: sparse.csr_matrix) -> IntArray:
    """
    Return an array of vertices in topological order.

    Parameters
    ----------
    A: sparse.csr_matrix
        CSR adjacency matrix of the directed acyclical graph.

    Returns
    -------
    sorted_vertices: np.ndarray of integer
    """
    return _topological_sort_by_dfs(_csr_to_adjacency(A))


@nb.njit
def _contract_vertices(A: AdjacencyMatrix, indices: IntArray) -> IntArray:
    vcolor = np.zeros(A.m, dtype=np.uint8)
    vcolor[indices] = 2

    edge_node_connectivity = np.empty((A.m, 2), dtype=np.int64)
    n_edge = 0

    S = np.empty(A.m, dtype=np.int64)
    size = 0

    for v in indices:
        size = 0
        size = push(S, v, size)
        while size > 0:
            u, size = pop(S, size)

            # Do not paint over the marked indices
            if vcolor[u] == 0:
                vcolor[u] == 1

            for n in neighbors(A, u):
                if (n == v) or (vcolor[n] == 1):
                    raise ValueError("The graph contains at least one cycle")
                elif vcolor[n] == 2:
                    edge_node_connectivity[n_edge, 0] = v
                    edge_node_connectivity[n_edge, 1] = n
                    n_edge += 1
                    size -= 1
                else:
                    size = push(S, n, size)

    return edge_node_connectivity[:n_edge]


def contract_vertices(A: sparse.csr_matrix, indices: IntArray) -> IntArray:
    """
    Contract vertices to the set defined by indices.

    Parameters
    ----------
    A: AdjacencyMatrix
        CSR adjacency matrix of the directed acyclical graph.
    indices: np.ndarray of integer
        Which vertices to preserve.

    Returns
    -------
    edge_node_connectivity: np.ndarray of shape (n_edge, 2)
        New edge_node_connectivity containing edges of the vertices contained
        in indices.
    """
    return _contract_vertices(_csr_to_adjacency(A), np.array(indices))


# Conversion between dense and sparse
# -----------------------------------
def _to_ij(conn: IntArray, invert: bool) -> Tuple[IntArray, IntArray]:
    n, m = conn.shape
    j = conn.ravel()
    valid = j != FILL_VALUE
    i = np.repeat(np.arange(n), m)[valid]
    j = j[valid]
    if invert:
        return j, i
    else:
        return i, j


def _to_sparse(conn: IntArray, invert: bool, sort_indices: bool) -> sparse.csr_matrix:
    i, j = _to_ij(conn, invert)
    coo_content = (j, (i, j))
    coo_matrix = sparse.coo_matrix(coo_content)
    csr_matrix = coo_matrix.tocsr()
    # Conversion to csr format results in a sorting of indices. We require
    # only sorting of i, not j, as this would e.g. mess up the order for
    # counter clockwise vertex orientation, e.g.
    if not sort_indices:
        order = np.argsort(i)
        csr_matrix.indices = j[order]
        csr_matrix.has_sorted_indices = False
    return csr_matrix


def ragged_index(n: int, m: int, m_per_row: IntArray) -> BoolArray:
    """
    Given an array of n rows by m columns, starting from left mark the values
    True such that the number of True values equals m_per_row.

    For example:

    n = 3
    m = 4
    m_per_row = np.array([1, 2, 3])

    Then the result of ragged_index(n, m, m_per_row) is:

    np.array([
        [True, False, False, False],
        [True, True, False, False],
        [True, True, True, False],
    ])

    This can be used as boolean index to set a variable number of values per
    row.
    """
    column_number = np.tile(np.arange(m), n).reshape((n, m))
    return (column_number.T < m_per_row).T


def to_sparse(conn: IntArray, sort_indices: bool = True) -> sparse.csr_matrix:
    return _to_sparse(conn, invert=False, sort_indices=sort_indices)


def to_dense(conn: SparseMatrix, n_columns: int = None) -> IntArray:
    n, _ = conn.shape
    m_per_row = conn.getnnz(axis=1)
    m = m_per_row.max()
    if n_columns is not None:
        if n_columns < m:
            raise ValueError(
                f"n_columns {n_columns} is too small for the data, requires {m}"
            )
        m = n_columns

    # Allocate 2D array and create a flat view of the dense connectivity
    dense_conn = np.empty((n, m), dtype=IntDType)
    flat_conn = dense_conn.ravel()
    if (n * m) == conn.nnz:
        # Shortcut if fill_value is not present, when all of same geom. type
        # e.g. all triangles or all quadrangles
        valid = slice(None)  # a[:] equals a[slice(None)]
    else:
        valid = ragged_index(n, m, m_per_row).ravel()
        flat_conn[~valid] = FILL_VALUE

    if isinstance(conn, sparse.csr_matrix):
        flat_conn[valid] = conn.indices
    elif isinstance(conn, sparse.coo_matrix):
        flat_conn[valid] = conn.col
    else:
        raise TypeError("Can only invert coo or csr matrix")
    return dense_conn


# Inverting connectivities
# ------------------------
def invert_dense_to_sparse(
    conn: IntArray, sort_indices: bool = True
) -> sparse.csr_matrix:
    return _to_sparse(conn, invert=True, sort_indices=sort_indices)


def invert_dense(conn: IntArray, sort_indices: bool = True) -> IntArray:
    sparse_inverted = _to_sparse(conn, invert=True, sort_indices=sort_indices)
    return to_dense(sparse_inverted)


def invert_sparse(conn: sparse.csr_matrix) -> sparse.csr_matrix:
    coo = conn.tocoo()
    j = coo.row
    i = coo.col
    coo_content = (j, (i, j))
    inverted = sparse.coo_matrix(coo_content)
    return inverted.tocsr()


def invert_sparse_to_dense(conn: sparse.csr_matrix) -> IntArray:
    inverted = invert_sparse(conn)
    return to_dense(inverted)


# Renumbering
# -----------
def _renumber(a: IntArray) -> IntArray:
    # Taken from https://github.com/scipy/scipy/blob/v1.7.1/scipy/stats/stats.py#L8631-L8737
    # (scipy is BSD-3-Clause License)
    arr = np.ravel(np.asarray(a))
    sorter = np.argsort(arr, kind="quicksort")
    inv = np.empty(sorter.size, dtype=IntDType)
    inv[sorter] = np.arange(sorter.size, dtype=IntDType)
    arr = arr[sorter]
    obs = np.r_[True, arr[1:] != arr[:-1]]
    dense = obs.cumsum()[inv] - 1
    return dense.reshape(a.shape)


def renumber(a: IntArray):
    valid = a != FILL_VALUE
    renumbered = np.full_like(a, FILL_VALUE)
    renumbered[valid] = _renumber(a[valid])
    return renumbered


def close_polygons(face_node_connectivity: IntArray) -> IntArray:
    # Wrap around and create closed polygon: put the first node at the end of the row
    # In case of fill values, replace all fill values
    n, m = face_node_connectivity.shape
    closed = np.full((n, m + 1), FILL_VALUE, dtype=IntDType)
    closed[:, :-1] = face_node_connectivity
    first_node = face_node_connectivity[:, 0]
    # Identify fill value, and replace by first node also
    isfill = closed == FILL_VALUE
    closed.ravel()[isfill.ravel()] = np.repeat(first_node, isfill.sum(axis=1))
    return closed, isfill


def reverse_orientation(face_node_connectivity: IntArray):
    # We cannot simply reverse the rows with [:, ::-1], since there may be fill
    # values present.
    reversed_orientation = face_node_connectivity.copy()
    in_reverse = face_node_connectivity[:, ::-1]
    in_reverse = in_reverse[in_reverse != FILL_VALUE]
    replace = face_node_connectivity != FILL_VALUE
    reversed_orientation[replace] = in_reverse
    return reversed_orientation


def counterclockwise(face_node_connectivity: IntArray, nodes: FloatArray) -> IntArray:
    # TODO: in case of "periodic grids", we need to wrap around the grid.
    closed, _ = close_polygons(face_node_connectivity)
    p = nodes[closed]
    dxy = np.diff(p, axis=1)
    reverse = cross2d(dxy[:, :-1], dxy[:, 1:]).sum(axis=1) < 0
    ccw = face_node_connectivity.copy()
    if reverse.any():
        ccw[reverse] = reverse_orientation(face_node_connectivity[reverse])
    return ccw


# Derived connectivities
# ----------------------
def boundary_node_connectivity(
    edge_face_connectivity: IntArray,
    edge_node_connectivity: IntArray,
) -> IntArray:
    """Is a subset of the edge_node_connectivity"""
    is_boundary = (edge_face_connectivity == FILL_VALUE).any(axis=1)
    return edge_node_connectivity[is_boundary]


def edge_connectivity(
    face_node_connectivity: IntArray,
    edge_node_connectivity=None,
) -> Tuple[IntArray, IntArray]:
    """Derive new edge_node_connectivity and face_edge_connectivity."""
    prior = edge_node_connectivity
    n, m = face_node_connectivity.shape
    # Close the polygons: [0 1 2 3] -> [0 1 2 3 0]
    closed, isfill = close_polygons(face_node_connectivity)
    # Allocate array for edge_node_connectivity: includes duplicate edges
    edge_node_connectivity = np.empty((n * m, 2), dtype=IntDType)
    edge_node_connectivity[:, 0] = closed[:, :-1].ravel()
    edge_node_connectivity[:, 1] = closed[:, 1:].ravel()
    # Cleanup: delete invalid edges (same node to same node)
    edge_node_connectivity = edge_node_connectivity[
        edge_node_connectivity[:, 0] != edge_node_connectivity[:, 1]
    ]
    # Now find the unique rows == unique edges
    edge_node_connectivity.sort(axis=1)
    edge_node_connectivity, inverse_indices = np.unique(
        ar=edge_node_connectivity, return_inverse=True, axis=0
    )
    inverse_indices = inverse_indices.ravel()

    if prior is not None:  # prior edge_node_connectivity exists
        unique, index = np.unique(np.sort(prior, axis=1), axis=0, return_index=True)
        # Check whether everything looks okay:
        if not np.array_equal(unique, edge_node_connectivity):
            raise ValueError(
                "Invalid edge_node_connectivity. Run .validate_edge_node_connectivity()."
            )
        inverse_indices = index[inverse_indices]
        edge_node_connectivity = prior

    # Create face_edge_connectivity
    face_edge_connectivity = np.full((n, m), FILL_VALUE, dtype=np.int64)
    isnode = ~isfill[:, :-1]
    face_edge_connectivity[isnode] = inverse_indices
    return edge_node_connectivity, face_edge_connectivity


def validate_edge_node_connectivity(
    face_node_connectivity: IntArray,
    edge_node_connectivity: IntArray,
) -> BoolArray:
    """
    * Is the edge defined by the face_node_connectivity?
    * Are the edges unique?
    """
    new, _ = edge_connectivity(face_node_connectivity)
    old = np.sort(edge_node_connectivity, axis=1)

    new1d = new.astype(np.int32).view(np.int64).ravel()
    old1d = old.astype(np.int32).view(np.int64).ravel()

    s = pd.Series(old1d)
    n_edge = len(new1d)
    n_old = s.nunique()
    if n_old < n_edge:
        raise ValueError(
            f"face_node_connectivity defines {n_edge} edges, but "
            f"edge_node_connectivity defines only {n_old} edges."
        )

    valid = np.isin(old1d, new1d) & (~s.duplicated().to_numpy())
    return valid


def face_face_connectivity(
    edge_face_connectivity: IntArray, n_face: int
) -> sparse.csr_matrix:
    """
    Derive face to face connectivity.

    An edge can be shared by two faces at most. If this is the case, they are
    neighbors.
    """
    i = edge_face_connectivity[:, 0]
    j = edge_face_connectivity[:, 1]
    is_connection = j != FILL_VALUE
    i = i[is_connection]
    j = j[is_connection]
    edge_index = np.arange(len(edge_face_connectivity))[is_connection]
    ij = np.concatenate([i, j])
    ji = np.concatenate([j, i])
    edge_index = np.concatenate([edge_index, edge_index])
    coo_content = (edge_index, (ij, ji))
    coo_matrix = sparse.coo_matrix(coo_content, shape=(n_face, n_face))
    return coo_matrix.tocsr()


def directed_node_node_connectivity(
    edge_node_connectivity: IntArray,
) -> sparse.csr_matrix:
    i = edge_node_connectivity[:, 0]
    j = edge_node_connectivity[:, 1]
    edge_index = np.arange(len(edge_node_connectivity))
    coo_content = (edge_index, (i, j))
    n = max(i.max(), j.max()) + 1
    coo_matrix = sparse.coo_matrix(coo_content, shape=(n, n))
    return coo_matrix.tocsr()


def node_node_connectivity(edge_node_connectivity: IntArray) -> sparse.csr_matrix:
    i = edge_node_connectivity[:, 0]
    j = edge_node_connectivity[:, 1]
    edge_index = np.arange(len(edge_node_connectivity))
    ij = np.concatenate([i, j])
    ji = np.concatenate([j, i])
    edge_index = np.concatenate([edge_index, edge_index])
    coo_content = (edge_index, (ij, ji))
    coo_matrix = sparse.coo_matrix(coo_content)
    return coo_matrix.tocsr()


def edge_edge_connectivity(
    edge_node_connectivity: IntArray,
    node_edge_connectivity: sparse.csr_matrix,
) -> sparse.csr_matrix:
    # - Flatten both nodes of the edge so we can index to the sparse matrix.
    # - Count the number of connected nodes for both the first and second node.
    # - Filter self -> self away.
    n_edge = len(edge_node_connectivity)
    node_index = edge_node_connectivity.ravel()
    j = node_edge_connectivity[node_index].indices
    n_connection = node_edge_connectivity.getnnz(axis=1)[node_index]
    # This way i comes pre-sorted in the COO constructor.
    i = np.repeat(np.arange(n_edge), n_connection.reshape((-1, 2)).sum(axis=1))
    is_connection = i != j
    data = np.repeat(node_index, n_connection)
    coo_content = (data[is_connection], (i[is_connection], j[is_connection]))
    coo_matrix = sparse.coo_matrix(coo_content)
    return coo_matrix.tocsr()


def directed_edge_edge_connectivity(
    edge_node_connectivity: IntArray,
    node_edge_connectivity: sparse.csr_matrix,
) -> sparse.csr_matrix:
    # - Get the second node of each edge.
    # - Find which edges are connected to this second node.
    # - This will include the edge we started with.
    # - Hence, make room for one more.
    # - Then filter self -> self away.
    n_edge = len(edge_node_connectivity)
    second_node = edge_node_connectivity[:, 1]
    n_downstream = node_edge_connectivity.getnnz(axis=1)[second_node]
    upstream_edges = np.repeat(np.arange(n_edge), n_downstream)
    downstream_edges = node_edge_connectivity[second_node].indices
    node_index = np.repeat(second_node, n_downstream)
    valid = downstream_edges != upstream_edges
    return sparse.csr_matrix(
        (node_index[valid], (upstream_edges[valid], downstream_edges[valid])),
        shape=(n_edge, n_edge),
    )


def structured_connectivity(active: IntArray) -> AdjacencyMatrix:
    nrow, ncol = active.shape
    nodes = np.arange(nrow * ncol).reshape(nrow, ncol)
    nodes[~active] = -1
    left = nodes[:, :-1].ravel()
    right = nodes[:, 1:].ravel()
    front = nodes[:-1].ravel()
    back = nodes[1:].ravel()
    # valid x connection
    valid = (left != -1) & (right != -1)
    left = left[valid]
    right = right[valid]
    # valid y connection
    valid = (front != -1) & (back != -1)
    front = front[valid]
    back = back[valid]
    i = renumber(np.concatenate([left, right, front, back]))
    j = renumber(np.concatenate([right, left, back, front]))
    coo_content = (j, (i, j))
    A = sparse.coo_matrix(coo_content).tocsr()
    n, m = A.shape
    return AdjacencyMatrix(A.indices, A.indptr, A.nnz, n, m)


def perimeter(
    face_node_connectivity: IntArray,
    node_x: FloatArray,
    node_y: FloatArray,
):
    nodes = np.column_stack([node_x, node_y])
    closed, _ = close_polygons(face_node_connectivity)
    coordinates = nodes[closed]
    # Shift coordinates to avoid precision loss
    xy0 = coordinates[:, 0]
    coordinates -= xy0[:, np.newaxis]
    dxy = np.diff(coordinates, axis=1)
    return np.linalg.norm(dxy, axis=-1).sum(axis=1)


def area_from_coordinates(
    coordinates: FloatArray,
) -> FloatArray:
    xy0 = coordinates[:, 0]
    a = coordinates[:, :-1] - xy0[:, np.newaxis]
    b = coordinates[:, 1:] - xy0[:, np.newaxis]
    determinant = cross2d(a, b)
    return 0.5 * abs(determinant.sum(axis=1))


def area(
    face_node_connectivity: IntArray,
    node_x: FloatArray,
    node_y: FloatArray,
) -> FloatArray:
    nodes = np.column_stack([node_x, node_y])
    closed, _ = close_polygons(face_node_connectivity)
    coordinates = nodes[closed]
    return area_from_coordinates(coordinates)


def centroids(
    face_node_connectivity: IntArray,
    node_x: FloatArray,
    node_y: FloatArray,
) -> FloatArray:
    n_face, n_max_node = face_node_connectivity.shape
    nodes = np.column_stack([node_x, node_y])
    # Check if it's fully triangular
    if n_max_node == 3:
        # Since it's triangular, computing the average of the vertices suffices
        coordinates = nodes[face_node_connectivity]
        return np.nanmean(coordinates, axis=1)
    else:
        # This is mathematically equivalent to triangulating, computing triangle centroids
        # and computing the area weighted average of those centroids
        centroid_coordinates = np.empty((n_face, 2), dtype=np.float64)
        closed, _ = close_polygons(face_node_connectivity)
        coordinates = nodes[closed]
        # Express coordinates relative to first node
        xy0 = coordinates[:, 0]
        a = coordinates[:, :-1] - xy0[:, np.newaxis]
        b = coordinates[:, 1:] - xy0[:, np.newaxis]
        c = a + b
        determinant = cross2d(a, b)
        area_weight = 1.0 / (3.0 * determinant.sum(axis=1))
        centroid_coordinates[:, 0] = area_weight * (c[..., 0] * determinant).sum(axis=1)
        centroid_coordinates[:, 1] = area_weight * (c[..., 1] * determinant).sum(axis=1)
        centroid_coordinates += xy0
        return centroid_coordinates


@nb.njit(cache=True, parallel=True)
def _circumcenters_triangle(xxx: FloatArray, yyy: FloatArray):
    """Numba should nicely fuse these operations."""
    a_x, b_x, c_x = xxx
    a_y, b_y, c_y = yyy
    D_inv = 0.5 / (
        a_y * c_x + b_y * a_x - b_y * c_x - a_y * b_x - c_y * a_x + c_y * b_x
    )
    x = ((a_x - c_x) * (a_x + c_x) + (a_y - c_y) * (a_y + c_y)) * (b_y - c_y) - (
        (b_x - c_x) * (b_x + c_x) + (b_y - c_y) * (b_y + c_y)
    ) * (a_y - c_y)
    y = ((b_x - c_x) * (b_x + c_x) + (b_y - c_y) * (b_y + c_y)) * (a_x - c_x) - (
        (a_x - c_x) * (a_x + c_x) + (a_y - c_y) * (a_y + c_y)
    ) * (b_x - c_x)
    return D_inv * x, D_inv * y


def circumcenters(
    face_node_connectivity: IntArray,
    node_x: FloatArray,
    node_y: FloatArray,
) -> FloatArray:
    # TODO: Skyum or Welzl implementation for polygons -- although it's
    # practical use is dubious?
    n_max_node = face_node_connectivity.shape[1]
    # Check if it's fully triangular
    if n_max_node == 3:
        xxx = node_x[face_node_connectivity.T]
        yyy = node_y[face_node_connectivity.T]
        x, y = _circumcenters_triangle(xxx, yyy)
    else:
        raise NotImplementedError(
            "Circumcenters are only supported for triangular grids"
        )
    return np.column_stack((x, y))


def _triangulate(i: IntArray, j: IntArray, n_triangle_per_row: IntArray) -> IntArray:
    n_triangle = n_triangle_per_row.sum()
    n_face = len(i)
    index_first = np.argwhere(np.diff(i, prepend=-1) != 0)
    index_second = index_first + 1
    index_last = np.argwhere(np.diff(i, append=-1) != 0)

    first = np.full(n_face, False)
    first[index_first] = True
    second = np.full(n_face, True) & ~first
    second[index_last] = False
    third = np.full(n_face, True) & ~first
    third[index_second] = False

    triangles = np.empty((n_triangle, 3), IntDType)
    triangles[:, 0] = np.repeat(j[first], n_triangle_per_row)
    triangles[:, 1] = j[second]
    triangles[:, 2] = j[third]
    return triangles


def triangulate_dense(face_node_connectivity: IntArray) -> Tuple[IntArray, IntArray]:
    n_face, n_max = face_node_connectivity.shape

    if n_max == 3:
        triangles = face_node_connectivity.copy()
        return triangles, np.arange(n_face)

    valid = face_node_connectivity != FILL_VALUE
    n_per_row = valid.sum(axis=1)
    n_triangle_per_row = n_per_row - 2
    i = np.repeat(np.arange(n_face), n_per_row)
    j = face_node_connectivity.ravel()[valid.ravel()]
    triangles = _triangulate(i, j, n_triangle_per_row)

    triangle_face_connectivity = np.repeat(
        np.arange(n_face), repeats=n_triangle_per_row
    )
    return triangles, triangle_face_connectivity


def triangulate_coo(
    face_node_connectivity: sparse.coo_matrix,
) -> Tuple[IntArray, IntArray]:
    ncol_per_row = face_node_connectivity.getnnz(axis=1)

    if ncol_per_row.max() == 3:
        triangles = face_node_connectivity.row.copy().reshape((-1, 3))
        return triangles, np.arange(len(triangles))

    n_triangle_per_row = ncol_per_row - 2
    i = face_node_connectivity.row
    j = face_node_connectivity.col
    triangles = _triangulate(i, j, n_triangle_per_row)

    n_face = face_node_connectivity.shape[0]
    triangle_face_connectivity = np.repeat(
        np.arange(n_face), repeats=n_triangle_per_row
    )
    return triangles, triangle_face_connectivity


def triangulate(face_node_connectivity) -> Tuple[IntArray, IntArray]:
    """
    Convert polygons into its constituent triangles.

    Triangulation runs from the first node of every face:

    * first, second, third,
    * first, third, fourth,
    * and so forth ...

    If the grid is already triangular, a copy is returned.

    Returns
    -------
    triangles: ndarray of integers with shape ``(n_triangle, 3)``
    triangle_face_connectivity: ndarray of integers with shape ``(n_triangle,)``
    """
    if isinstance(face_node_connectivity, IntArray):
        return triangulate_dense(face_node_connectivity)
    elif isinstance(face_node_connectivity, sparse.coo_matrix):
        return triangulate_coo(face_node_connectivity)
    else:
        raise TypeError("connectivity must be ndarray or sparse matrix")


def _mutate(
    output: BoolArray, i: IntArray, j: IntArray, value: bool, mask: BoolArray
) -> None:
    a = output[i]
    b = output[j]
    mutate = a != b
    output[i[mutate]] = value
    output[j[mutate]] = value
    if mask is not None:
        output[mask] = not value
    return None


def _binary_iterate(
    connectivity: sparse.csr_matrix,
    input: BoolArray,
    value: bool,
    iterations: int,
    mask: BoolArray,
    exterior: IntArray,
    border_value: bool,
) -> BoolArray:
    if input.dtype != np.bool_:
        raise TypeError("input dtype should be bool")

    ndim = input.ndim
    if ndim != 1:
        raise ValueError(
            "Binary operations are only supported for a single (face)"
            f"dimension. Found {ndim} dimensions."
        )

    coo = connectivity.tocoo()
    i = coo.row
    j = coo.col
    output = input.copy()

    # First iteration, mutate borders.
    _mutate(output, i, j, value, mask)
    if exterior is not None and value == border_value:
        output[exterior] = value

    # Subsequent iterations, disregard border.
    for _ in range(iterations - 1):
        _mutate(output, i, j, value, mask)

    return output


def binary_erosion(
    connectivity: sparse.csr_matrix,
    input: BoolArray,
    iterations: int = 1,
    mask: BoolArray | None = None,
    exterior: IntArray | None = None,
    border_value: bool | None = False,
) -> BoolArray:
    """By default, erodes inwards from the exterior."""
    return _binary_iterate(
        connectivity=connectivity,
        input=input,
        value=False,
        iterations=iterations,
        mask=mask,
        exterior=exterior,
        border_value=border_value,
    )


def binary_dilation(
    connectivity: sparse.csr_matrix,
    input: BoolArray,
    iterations: int = 1,
    mask: BoolArray | None = None,
    exterior: IntArray | None = None,
    border_value: bool | None = False,
) -> BoolArray:
    """By default, does not dilate inward from the exterior."""
    return _binary_iterate(
        connectivity=connectivity,
        input=input,
        value=True,
        iterations=iterations,
        mask=mask,
        exterior=exterior,
        border_value=border_value,
    )



"""
This module is a subset of the testing functions in xarray.tests.test_plot. The
reason is that we use a number of plot utils that are private to xarray.

Additionally, the _importorskip has been copied.

We heavily discourage editing this file. Any update should only consist of
copying updated parts of the xarray module.

Xarray is licensed under Apache License 2.0:
https://github.com/pydata/xarray/blob/main/LICENSE
"""

from __future__ import annotations

import contextlib
import importlib
import math

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal  # noqa: F401
from packaging.version import Version
from xarray import DataArray

from xugrid.plot.utils import (
    _color_palette,
    _determine_cmap_params,
    _maybe_gca,
    get_axis,
    label_from_attrs,
)

# import mpl and change the backend before other mpl imports
try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ImportError:
    pass


def _importorskip(
    modname: str, minversion: str | None = None
) -> tuple[bool, pytest.MarkDecorator]:
    try:
        mod = importlib.import_module(modname)
        has = True
        if minversion is not None:
            if Version(mod.__version__) < Version(minversion):
                raise ImportError("Minimum version not satisfied")
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f"requires {modname}")
    return has, func


has_matplotlib, requires_matplotlib = _importorskip("matplotlib")


@contextlib.contextmanager
def figure_context(*args, **kwargs):
    """Context manager which autocloses a figure (even if the test failed)"""

    try:
        yield None
    finally:
        plt.close("all")


def easy_array(shape, start=0, stop=1):
    """
    Make an array with desired shape using np.linspace

    shape is a tuple like (2, 3)
    """
    a = np.linspace(start, stop, num=math.prod(shape))
    return a.reshape(shape)


@requires_matplotlib
class PlotTestCase:
    @pytest.fixture(autouse=True)
    def setup(self):
        yield
        # Remove all matplotlib figures
        plt.close("all")

    def pass_in_axis(self, plotmethod, subplot_kw=None):
        fig, axs = plt.subplots(ncols=2, subplot_kw=subplot_kw)
        plotmethod(ax=axs[0])
        assert axs[0].has_data()

    def imshow_called(self, plotmethod):
        plotmethod()
        images = plt.gca().findobj(mpl.image.AxesImage)
        return len(images) > 0

    def contourf_called(self, plotmethod):
        plotmethod()

        # Compatible with mpl before (PathCollection) and after (QuadContourSet) 3.8
        def matchfunc(x):
            return isinstance(
                x, (mpl.collections.PathCollection, mpl.contour.QuadContourSet)
            )

        paths = plt.gca().findobj(matchfunc)
        return len(paths) > 0


@requires_matplotlib
class TestDiscreteColorMap:
    @pytest.fixture(autouse=True)
    def setUp(self):
        x = np.arange(start=0, stop=10, step=2)
        y = np.arange(start=9, stop=-7, step=-3)
        xy = np.dstack(np.meshgrid(x, y))
        distance = np.linalg.norm(xy, axis=2)
        self.darray = DataArray(distance, list(zip(("y", "x"), (y, x))))
        self.data_min = distance.min()
        self.data_max = distance.max()
        yield
        # Remove all matplotlib figures
        plt.close("all")

    def test_recover_from_seaborn_jet_exception(self) -> None:
        pal = _color_palette("jet", 4)
        assert isinstance(pal, np.ndarray)
        assert len(pal) == 4


@requires_matplotlib
class TestDetermineCmapParams:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.data = np.linspace(0, 1, num=100)

    def test_robust(self) -> None:
        cmap_params = _determine_cmap_params(self.data, robust=True)
        assert cmap_params["vmin"] == np.percentile(self.data, 2)
        assert cmap_params["vmax"] == np.percentile(self.data, 98)
        assert cmap_params["cmap"] == "viridis"
        assert cmap_params["extend"] == "both"
        assert cmap_params["levels"] is None
        assert cmap_params["norm"] is None

    def test_center(self) -> None:
        cmap_params = _determine_cmap_params(self.data, center=0.5)
        assert cmap_params["vmax"] - 0.5 == 0.5 - cmap_params["vmin"]
        assert cmap_params["cmap"] == "RdBu_r"
        assert cmap_params["extend"] == "neither"
        assert cmap_params["levels"] is None
        assert cmap_params["norm"] is None

    # EDIT: removed tests for xr.set_options.
    def test_nan_inf_are_ignored(self) -> None:
        cmap_params1 = _determine_cmap_params(self.data)
        data = self.data
        data[50:55] = np.nan
        data[56:60] = np.inf
        cmap_params2 = _determine_cmap_params(data)
        assert cmap_params1["vmin"] == cmap_params2["vmin"]
        assert cmap_params1["vmax"] == cmap_params2["vmax"]

    def test_integer_levels(self) -> None:
        data = self.data + 1

        # default is to cover full data range but with no guarantee on Nlevels
        for level in np.arange(2, 10, dtype=int):
            cmap_params = _determine_cmap_params(data, levels=level)
            assert cmap_params["vmin"] is None
            assert cmap_params["vmax"] is None
            assert cmap_params["norm"].vmin == cmap_params["levels"][0]
            assert cmap_params["norm"].vmax == cmap_params["levels"][-1]
            assert cmap_params["extend"] == "neither"

        # with min max we are more strict
        cmap_params = _determine_cmap_params(
            data, levels=5, vmin=0, vmax=5, cmap="Blues"
        )
        assert cmap_params["vmin"] is None
        assert cmap_params["vmax"] is None
        assert cmap_params["norm"].vmin == 0
        assert cmap_params["norm"].vmax == 5
        assert cmap_params["norm"].vmin == cmap_params["levels"][0]
        assert cmap_params["norm"].vmax == cmap_params["levels"][-1]
        assert cmap_params["cmap"].name == "Blues"
        assert cmap_params["extend"] == "neither"
        assert cmap_params["cmap"].N == 4
        assert cmap_params["norm"].N == 5

        cmap_params = _determine_cmap_params(data, levels=5, vmin=0.5, vmax=1.5)
        assert cmap_params["cmap"].name == "viridis"
        assert cmap_params["extend"] == "max"

        cmap_params = _determine_cmap_params(data, levels=5, vmin=1.5)
        assert cmap_params["cmap"].name == "viridis"
        assert cmap_params["extend"] == "min"

        cmap_params = _determine_cmap_params(data, levels=5, vmin=1.3, vmax=1.5)
        assert cmap_params["cmap"].name == "viridis"
        assert cmap_params["extend"] == "both"

    def test_list_levels(self) -> None:
        data = self.data + 1

        orig_levels = [0, 1, 2, 3, 4, 5]
        # vmin and vmax should be ignored if levels are explicitly provided
        cmap_params = _determine_cmap_params(data, levels=orig_levels, vmin=0, vmax=3)
        assert cmap_params["vmin"] is None
        assert cmap_params["vmax"] is None
        assert cmap_params["norm"].vmin == 0
        assert cmap_params["norm"].vmax == 5
        assert cmap_params["cmap"].N == 5
        assert cmap_params["norm"].N == 6

        for wrap_levels in [list, np.array, pd.Index, DataArray]:
            cmap_params = _determine_cmap_params(data, levels=wrap_levels(orig_levels))
            assert_array_equal(cmap_params["levels"], orig_levels)

    def test_divergentcontrol(self) -> None:
        neg = self.data - 0.1
        pos = self.data

        # Default with positive data will be a normal cmap
        cmap_params = _determine_cmap_params(pos)
        assert cmap_params["vmin"] == 0
        assert cmap_params["vmax"] == 1
        assert cmap_params["cmap"] == "viridis"

        # Default with negative data will be a divergent cmap
        cmap_params = _determine_cmap_params(neg)
        assert cmap_params["vmin"] == -0.9
        assert cmap_params["vmax"] == 0.9
        assert cmap_params["cmap"] == "RdBu_r"

        # Setting vmin or vmax should prevent this only if center is false
        cmap_params = _determine_cmap_params(neg, vmin=-0.1, center=False)
        assert cmap_params["vmin"] == -0.1
        assert cmap_params["vmax"] == 0.9
        assert cmap_params["cmap"] == "viridis"
        cmap_params = _determine_cmap_params(neg, vmax=0.5, center=False)
        assert cmap_params["vmin"] == -0.1
        assert cmap_params["vmax"] == 0.5
        assert cmap_params["cmap"] == "viridis"

        # Setting center=False too
        cmap_params = _determine_cmap_params(neg, center=False)
        assert cmap_params["vmin"] == -0.1
        assert cmap_params["vmax"] == 0.9
        assert cmap_params["cmap"] == "viridis"

        # However, I should still be able to set center and have a div cmap
        cmap_params = _determine_cmap_params(neg, center=0)
        assert cmap_params["vmin"] == -0.9
        assert cmap_params["vmax"] == 0.9
        assert cmap_params["cmap"] == "RdBu_r"

        # Setting vmin or vmax alone will force symmetric bounds around center
        cmap_params = _determine_cmap_params(neg, vmin=-0.1)
        assert cmap_params["vmin"] == -0.1
        assert cmap_params["vmax"] == 0.1
        assert cmap_params["cmap"] == "RdBu_r"
        cmap_params = _determine_cmap_params(neg, vmax=0.5)
        assert cmap_params["vmin"] == -0.5
        assert cmap_params["vmax"] == 0.5
        assert cmap_params["cmap"] == "RdBu_r"
        cmap_params = _determine_cmap_params(neg, vmax=0.6, center=0.1)
        assert cmap_params["vmin"] == -0.4
        assert cmap_params["vmax"] == 0.6
        assert cmap_params["cmap"] == "RdBu_r"

        # But this is only true if vmin or vmax are negative
        cmap_params = _determine_cmap_params(pos, vmin=-0.1)
        assert cmap_params["vmin"] == -0.1
        assert cmap_params["vmax"] == 0.1
        assert cmap_params["cmap"] == "RdBu_r"
        cmap_params = _determine_cmap_params(pos, vmin=0.1)
        assert cmap_params["vmin"] == 0.1
        assert cmap_params["vmax"] == 1
        assert cmap_params["cmap"] == "viridis"
        cmap_params = _determine_cmap_params(pos, vmax=0.5)
        assert cmap_params["vmin"] == 0
        assert cmap_params["vmax"] == 0.5
        assert cmap_params["cmap"] == "viridis"

        # If both vmin and vmax are provided, output is non-divergent
        cmap_params = _determine_cmap_params(neg, vmin=-0.2, vmax=0.6)
        assert cmap_params["vmin"] == -0.2
        assert cmap_params["vmax"] == 0.6
        assert cmap_params["cmap"] == "viridis"

        # regression test for GH3524
        # infer diverging colormap from divergent levels
        cmap_params = _determine_cmap_params(pos, levels=[-0.1, 0, 1])
        # specifying levels makes cmap a Colormap object
        assert cmap_params["cmap"].name == "RdBu_r"

    def test_norm_sets_vmin_vmax(self) -> None:
        vmin = self.data.min()
        vmax = self.data.max()

        for norm, extend, levels in zip(
            [
                mpl.colors.Normalize(),
                mpl.colors.Normalize(),
                mpl.colors.Normalize(vmin + 0.1, vmax - 0.1),
                mpl.colors.Normalize(None, vmax - 0.1),
                mpl.colors.Normalize(vmin + 0.1, None),
            ],
            ["neither", "neither", "both", "max", "min"],
            [7, None, None, None, None],
        ):
            test_min = vmin if norm.vmin is None else norm.vmin
            test_max = vmax if norm.vmax is None else norm.vmax

            cmap_params = _determine_cmap_params(self.data, norm=norm, levels=levels)
            assert cmap_params["vmin"] is None
            assert cmap_params["vmax"] is None
            assert cmap_params["norm"].vmin == test_min
            assert cmap_params["norm"].vmax == test_max
            assert cmap_params["extend"] == extend
            assert cmap_params["norm"] == norm


@requires_matplotlib
def test_get_axis_current() -> None:
    with figure_context():
        _, ax = plt.subplots()
        out_ax = get_axis()
        assert ax is out_ax


@requires_matplotlib
def test_maybe_gca() -> None:
    with figure_context():
        ax = _maybe_gca(aspect=1)

        assert isinstance(ax, mpl.axes.Axes)
        assert ax.get_aspect() == 1

    with figure_context():
        # create figure without axes
        plt.figure()
        ax = _maybe_gca(aspect=1)

        assert isinstance(ax, mpl.axes.Axes)
        assert ax.get_aspect() == 1

    with figure_context():
        existing_axes = plt.axes()
        ax = _maybe_gca(aspect=1)

        # re-uses the existing axes
        assert existing_axes == ax
        # kwargs are ignored when reusing axes
        assert ax.get_aspect() == "auto"


class TestPlot(PlotTestCase):
    @pytest.fixture(autouse=True)
    def setup_array(self) -> None:
        self.darray = DataArray(easy_array((2, 3, 4)))

    def test_accessor(self) -> None:
        from xarray.plot.accessor import DataArrayPlotAccessor

        assert DataArray.plot is DataArrayPlotAccessor
        assert isinstance(self.darray.plot, DataArrayPlotAccessor)

    def test_label_from_attrs(self) -> None:
        da = self.darray.copy()
        assert "" == label_from_attrs(da)

        da.name = 0
        assert "0" == label_from_attrs(da)

        da.name = "a"
        da.attrs["units"] = "a_units"
        da.attrs["long_name"] = "a_long_name"
        da.attrs["standard_name"] = "a_standard_name"
        assert "a_long_name [a_units]" == label_from_attrs(da)

        da.attrs.pop("long_name")
        assert "a_standard_name [a_units]" == label_from_attrs(da)
        da.attrs.pop("units")
        assert "a_standard_name" == label_from_attrs(da)

        da.attrs["units"] = "a_units"
        da.attrs.pop("standard_name")
        assert "a [a_units]" == label_from_attrs(da)

        da.attrs.pop("units")
        assert "a" == label_from_attrs(da)

        # Latex strings can be longer without needing a new line:
        long_latex_name = r"$Ra_s = \mathrm{mean}(\epsilon_k) / \mu M^2_\infty$"
        da.attrs = {"long_name": long_latex_name}
        assert label_from_attrs(da) == long_latex_name



"""
Conversion from and to other data structures:

* GIS vector data (e.g. geopackage, shapefile)
* Structured data (e.g. rasters)

"""

import warnings
from typing import Tuple, Union

import numpy as np
import xarray as xr

from xugrid.constants import (
    FILL_VALUE,
    BoolArray,
    FloatArray,
    IntArray,
    IntDType,
    LineArray,
    MissingOptionalModule,
    PointArray,
    PolygonArray,
)
from xugrid.ugrid.connectivity import cross2d, ragged_index
from xugrid.ugrid.ugrid1d import Ugrid1d
from xugrid.ugrid.ugrid2d import Ugrid2d

try:
    import shapely
except ImportError:
    shapely = MissingOptionalModule("shapely")


def contiguous_xy(xy: FloatArray) -> Tuple[FloatArray, FloatArray]:
    x, y = [np.ascontiguousarray(a) for a in xy.T]
    return x, y


def nodes_to_points(x: FloatArray, y: FloatArray) -> PointArray:
    return shapely.points(x, y)


def edges_to_linestrings(
    x: FloatArray, y: FloatArray, edge_node_connectivity: IntArray
) -> LineArray:
    c = edge_node_connectivity.ravel()
    xy = np.column_stack((x[c], y[c]))
    i = np.repeat(np.arange(len(edge_node_connectivity)), 2)
    return shapely.linestrings(xy, indices=i)


def faces_to_polygons(
    x: FloatArray, y: FloatArray, face_node_connectivity: IntArray
) -> PolygonArray:
    is_data = face_node_connectivity != FILL_VALUE
    m_per_row = is_data.sum(axis=1)
    i = np.repeat(np.arange(len(face_node_connectivity)), m_per_row)
    c = face_node_connectivity.ravel()[is_data.ravel()]
    xy = np.column_stack((x[c], y[c]))
    rings = shapely.linearrings(xy, indices=i)
    return shapely.polygons(rings)


def points_to_nodes(points: PointArray) -> Tuple[FloatArray, FloatArray]:
    return contiguous_xy(shapely.get_coordinates(points))


def linestrings_to_edges(edges: LineArray) -> Tuple[FloatArray, FloatArray, IntArray]:
    xy, index = shapely.get_coordinates(edges, return_index=True)
    linear_index = np.arange(index.size)
    segments = np.column_stack([linear_index[:-1], linear_index[1:]])
    segments = segments[np.diff(index) == 0]
    unique, inverse = np.unique(xy, return_inverse=True, axis=0)
    inverse = inverse.ravel()
    segments = inverse[segments]
    x, y = contiguous_xy(unique)
    return x, y, segments


def _remove_last_vertex(xy: FloatArray, indices: IntArray):
    """
    GEOS polygons are always closed: the first and last vertex are the same.
    UGRID faces are by definition closed. So we'll have to remove the last,
    repeated, vertex of every polygon here.
    """
    no_repeats = np.diff(indices, append=-1) == 0
    xy = xy[no_repeats]
    indices = indices[no_repeats]
    return xy, indices


def polygons_to_faces(
    polygons: PolygonArray,
) -> Tuple[FloatArray, FloatArray, IntArray, int]:
    xy, indices = _remove_last_vertex(
        *shapely.get_coordinates(polygons, return_index=True)
    )
    unique, inverse = np.unique(xy, axis=0, return_inverse=True)
    inverse = inverse.ravel()
    n = len(polygons)
    m_per_row = np.bincount(indices)
    m = m_per_row.max()
    # Allocate 2D array and create a flat view of the dense connectivity
    conn = np.empty((n, m), dtype=IntDType)
    flat_conn = conn.ravel()
    if (n * m) == indices.size:
        # Shortcut if fill_value is not present, when all of same geom. type
        # e.g. all triangles or all quadrangles
        valid = slice(None)  # a[:] equals a[slice(None)]
    else:
        valid = ragged_index(n, m, m_per_row).ravel()
        flat_conn[~valid] = FILL_VALUE
    flat_conn[valid] = inverse
    x, y = contiguous_xy(unique)
    return x, y, conn


def _scalar_spacing(coords, spacing):
    dim = coords.dims[0]
    diff = coords.diff(dim)
    spacing_value = abs(spacing.item())
    if not np.allclose(
        abs(diff.to_numpy()), spacing_value, atol=abs(1.0e-4 * spacing.item())
    ):
        raise ValueError(
            f"spacing of {coords.name} does not match value of {spacing.name}"
        )
    halfdiff = xr.full_like(coords, 0.5 * spacing_value)
    return halfdiff


def _array_spacing(coords, spacing):
    if coords.size != spacing.size:
        raise ValueError(f"size of {coords.name} does not match size of {spacing.name}")
    halfdiff = 0.5 * abs(spacing)
    return halfdiff


def _is_monotonic_and_increasing(coord, axis=0) -> bool:
    """
    Test if monotonic and retun whether increasing along axis.
    Raises error if not monotonic.

    Copied and slightly adapted from xarray.utils.

    >>> _is_monotonic(np.array([0, 1, 2]))
    True
    >>> _is_monotonic(np.array([2, 1, 0]))
    True
    >>> _is_monotonic(np.array([0, 2, 1]))
    False
    """
    coord = np.asarray(coord)
    n = coord.shape[axis]
    delta_pos = coord.take(np.arange(1, n), axis=axis) >= coord.take(
        np.arange(0, n - 1), axis=axis
    )
    delta_neg = coord.take(np.arange(1, n), axis=axis) <= coord.take(
        np.arange(0, n - 1), axis=axis
    )
    if np.all(delta_pos):
        return True
    elif np.all(delta_neg):
        return False
    else:
        raise ValueError("The input coordinate is not monotonic.")


def infer_interval_breaks(coord, axis: int = 0, check_monotonic: bool = False):
    """
    Infer intervals from cell center coordinates.

    Copied and adapted from xarray.utils.

    >>> _infer_interval_breaks(np.arange(5))
    array([-0.5,  0.5,  1.5,  2.5,  3.5,  4.5])
    >>> _infer_interval_breaks([[0, 1], [3, 4]], axis=1)
    array([[-0.5,  0.5,  1.5],
           [ 2.5,  3.5,  4.5]])
    """
    coord = np.asarray(coord)
    if check_monotonic:
        _is_monotonic_and_increasing(coord, axis=axis)

    deltas = 0.5 * np.diff(coord, axis=axis)
    if deltas.size == 0:
        deltas = np.array(0.0)

    first = np.take(coord, [0], axis=axis) - np.take(deltas, [0], axis=axis)
    last = np.take(coord, [-1], axis=axis) + np.take(deltas, [-1], axis=axis)
    trim_last = tuple(
        slice(None, -1) if n == axis else slice(None) for n in range(coord.ndim)
    )
    interval_breaks = np.concatenate(
        [first, coord[trim_last] + deltas, last], axis=axis
    )
    return interval_breaks


def infer_interval_breaks1d(
    obj: Union[xr.DataArray, xr.Dataset],
    var: str,
) -> np.ndarray:
    """
    Infer the breaks for 1D coordinates.

        * For non-equidistant grids, taking half of each cell size result in
          wrong answers.
        * Cell size may be provided by a dx or dy attribute instead.
        * We also want to take 1-row or 1 column topologies into account.
    """
    coord = obj[var]
    spacing_name = f"d{var}"

    # Spacing name is provided:
    if spacing_name in obj.coords:
        spacing = obj[spacing_name]
        if spacing.ndim > 1:
            raise NotImplementedError(
                f"More than one dimension in spacing variable: {spacing_name}"
            )
        if spacing.shape in ((), (1,)):
            halfdiff = _scalar_spacing(coord, spacing)
        else:
            halfdiff = _array_spacing(coord, spacing)

        # Now check if monotonic and take orientation into account.
        if _is_monotonic_and_increasing(coord):
            intervals = np.insert(coord + halfdiff, 0, coord[0] - halfdiff[0])
        else:
            intervals = np.insert(coord - halfdiff, 0, coord[0] + halfdiff[0])

    # Implicit spacing, infer from coordinates instead:
    else:
        if coord.size == 1:
            raise ValueError(
                f"Cannot derive spacing of 1-sized coordinate: {var} \n"
                f"Assign a d{var} variable with spacing instead."
            )
        intervals = infer_interval_breaks(coord.to_numpy(), check_monotonic=True)

    return intervals


def infer_xy_coords(obj):
    # First check names, then check whether CF roles are specified.
    x = None
    y = None
    if "x" in obj.dims and "y" in obj.dims:
        x, y = "x", "y"
    elif "longitude" in obj.dims and "latitude" in obj.dims:
        x, y = "longitude", "latitude"
    else:
        for name, da in obj.coords.items():
            # Only 1D dimensions are allowed.
            if da.ndim != 1:
                continue

            attrs = da.attrs
            axis = attrs.get("axis", "").lower()
            stdname = attrs.get("standard_name", "").lower()
            if axis == "x" or stdname in ("longitude", "projection_x_coordinate"):
                x = name
            elif axis == "y" or stdname in ("latitude", "projection_y_coordinate"):
                y = name

    return x, y


def bounds1d_to_vertices(bounds: np.ndarray):
    diff = np.diff(bounds, axis=0)
    ascending = (diff >= 0.0).all()
    descending = (diff <= 0.0).all()
    if ascending:
        vertices = np.concatenate((bounds[:, 0], bounds[-1:, 1]))
    elif descending:
        vertices = np.concatenate((bounds[:, 1], bounds[-1:, 0]))
    else:
        raise ValueError("Bounds are not monotonic ascending or monotonic descending")
    return vertices


def quad_area(coordinates: FloatArray):
    # Subtly different from connectivity.area_from_coordinates
    # Due to lexsort (for repeated values), coordinates are not in CCW order.
    # We might get conflicting sign on the determinant; we always get two triangles,
    # so call abs() before summing.
    xy0 = coordinates[:, 0]
    a = coordinates[:, :-1] - xy0[:, np.newaxis]
    b = coordinates[:, 1:] - xy0[:, np.newaxis]
    determinant = cross2d(a, b)
    return 0.5 * abs(determinant).sum(axis=1)


def bounds2d_to_topology2d(
    x_bounds: np.ndarray, y_bounds: np.ndarray
) -> Tuple[FloatArray, BoolArray]:
    x = x_bounds.reshape(-1, 4)
    y = y_bounds.reshape(-1, 4)
    # Make sure repeated nodes are consecutive so we can find them later.
    # lexsort along axis 1.
    sorter = np.lexsort((y, x))
    face_node_coordinates = np.stack(
        (
            np.take_along_axis(x, sorter, axis=1),
            np.take_along_axis(y, sorter, axis=1),
        ),
        axis=-1,
    )

    # Check whether all coordinates form valid UGRID topologies.
    # We can only maintain triangles and quadrangles.
    # We also have to discard collinear triangles or quadrangles.
    n_unique = (
        (face_node_coordinates != np.roll(face_node_coordinates, 1, axis=1))
        .any(axis=-1)
        .sum(axis=1)
    )
    valid = (n_unique >= 3) & (quad_area(face_node_coordinates) > 0)
    if not valid.all():
        warnings.warn(
            "A UGRID2D face requires at least three unique non-collinear vertices.\n"
            f"Your structured bounds contain {len(valid) - valid.sum()} invalid faces.\n"
            "These will be omitted from the Ugrid2d topology.",
            UserWarning,
            stacklevel=2,
        )
    # Also check for NaNs.
    index = np.isfinite(face_node_coordinates.reshape(-1, 8)).all(axis=-1) & valid
    face_node_coordinates = face_node_coordinates[index]

    # Guarantee counterclockwise orientation.
    face_centroids = np.mean(face_node_coordinates, axis=1)
    dx = face_node_coordinates[..., 0] - face_centroids[:, np.newaxis, 0]
    dy = face_node_coordinates[..., 1] - face_centroids[:, np.newaxis, 1]
    angle = np.arctan2(dy, dx)
    # When nodes are repeated, make sure the repeated node ends up as last so we can
    # construct the face_node_connectivity directly from it. We do so by inserting
    # an angle of np.inf for the repeated nodes.
    angle[:, 1:][angle[:, 1:] == angle[:, :-1]] = np.inf
    counterclockwise = np.argsort(angle, axis=1)
    face_node_coordinates = np.take_along_axis(
        face_node_coordinates, counterclockwise[..., None], axis=1
    )
    # TODO: this assumes bounds align exactly. Do we need a tolerance?
    xy, inverse = np.unique(
        face_node_coordinates.reshape((-1, 2)), return_inverse=True, axis=0
    )
    face_node_connectivity = inverse.reshape((-1, 4))
    # For triangles, set the last node to the fill value of -1.
    face_node_connectivity[n_unique[index] == 3, -1] = -1
    return *xy.T, face_node_connectivity, index


def grid_from_geodataframe(geodataframe: "geopandas.GeoDataFrame"):  # type: ignore # noqa
    import geopandas as gpd

    gdf = geodataframe
    if not isinstance(gdf, gpd.GeoDataFrame):
        raise TypeError(
            f"Cannot convert a {type(gdf).__name__}, expected a GeoDataFrame"
        )

    geom_types = gdf.geom_type.unique()
    if len(geom_types) == 0:
        raise ValueError("geodataframe contains no geometry")
    elif len(geom_types) > 1:
        message = ", ".join(geom_types)
        raise ValueError(f"Multiple geometry types detected: {message}")

    geom_type = geom_types[0]
    if geom_type == "LineString":
        grid = Ugrid1d.from_geodataframe(gdf)
    elif geom_type == "Polygon":
        grid = Ugrid2d.from_geodataframe(gdf)
    else:
        raise ValueError(
            f"Invalid geometry type: {geom_type}. Expected Linestring or Polygon."
        )
    return grid


def grid_from_dataset(dataset: xr.Dataset, topology: str):
    topodim = dataset[topology].attrs["topology_dimension"]
    if topodim == 1:
        return Ugrid1d.from_dataset(dataset, topology)
    elif topodim == 2:
        return Ugrid2d.from_dataset(dataset, topology)
    elif topodim == 3:
        raise NotImplementedError
    else:
        raise ValueError(f"Invalid topology dimension: {topodim}")



import importlib

import numpy as np
import pooch
import xarray as xr

import xugrid

REGISTRY = pooch.create(
    path=pooch.os_cache("xugrid"),
    base_url="https://github.com/deltares/xugrid/raw/main/data/",
    version=None,
    version_dev="main",
    env="XUGRID_DATA_DIR",
)
REGISTRY.load_registry(importlib.resources.files("xugrid.data") / "registry.txt")


def xoxo():
    """Fetch a simple two part synthetic unstructured grid topology."""
    fname_vertices = REGISTRY.fetch("xoxo_vertices.txt")
    fname_triangles = REGISTRY.fetch("xoxo_triangles.txt")
    vertices = np.loadtxt(fname_vertices, dtype=float)
    triangles = np.loadtxt(fname_triangles, dtype=int)
    grid = xugrid.Ugrid2d(
        node_x=vertices[:, 0],
        node_y=vertices[:, 1],
        fill_value=-1,
        face_node_connectivity=triangles,
    )
    return grid


def adh_san_diego(xarray=False):
    """Fetch time varying output of a hydraulic simulation."""
    fname = REGISTRY.fetch("ADH_SanDiego.nc")
    ds = xr.open_dataset(fname)
    ds["node_x"].attrs["standard_name"] = "projection_x_coordinate"
    ds["node_y"].attrs["standard_name"] = "projection_y_coordinate"
    if xarray:
        return ds
    else:
        grid = xugrid.Ugrid2d.from_dataset(ds)
        return xugrid.UgridDataset(ds, grid)


def elevation_nl(xarray=False):
    """Fetch surface elevation dataset for the Netherlands."""
    fname = REGISTRY.fetch("elevation_nl.nc")
    ds = xr.open_dataset(fname)
    ds["mesh2d_node_x"].attrs["standard_name"] = "projection_x_coordinate"
    ds["mesh2d_node_y"].attrs["standard_name"] = "projection_y_coordinate"
    ds["mesh2d_face_x"].attrs["standard_name"] = "projection_x_coordinate"
    ds["mesh2d_face_y"].attrs["standard_name"] = "projection_y_coordinate"
    if xarray:
        return ds
    else:
        grid = xugrid.Ugrid2d.from_dataset(ds)
        return xugrid.UgridDataArray(ds["elevation"], grid)


def provinces_nl():
    """Fetch provinces polygons for the Netherlands."""
    import geopandas as gpd

    fname = REGISTRY.fetch("provinces-nl.geojson")
    gdf = gpd.read_file(fname)
    return gdf


def hydamo_network():
    """Fetch some surface water data for the Netherlands."""
    import geopandas as gpd
    import pandas as pd

    def read_wkt_csv(path):
        df = pd.read_csv(path)
        return gpd.GeoDataFrame(
            data=df,
            geometry=gpd.GeoSeries.from_wkt(df["geometry"]),
        )

    objects_fname = REGISTRY.fetch("hydamo_objects.csv")
    points_fname = REGISTRY.fetch("hydamo_points.csv")
    profiles_fname = REGISTRY.fetch("hydamo_profiles.csv")

    return (
        read_wkt_csv(objects_fname),
        read_wkt_csv(points_fname),
        read_wkt_csv(profiles_fname),
    )



import geopandas as gpd
import numpy as np
import pytest
import xarray as xr

import xugrid as xu


@pytest.fixture(scope="function")
def grid():
    """Three by three squares"""
    x = np.arange(0.0, 4.0)
    y = np.arange(0.0, 4.0)
    node_y, node_x = [a.ravel() for a in np.meshgrid(y, x, indexing="ij")]
    nx = ny = 3
    # Define the first vertex of every face, v.
    v = (np.add.outer(np.arange(nx), nx * np.arange(ny)) + np.arange(ny)).T.ravel()
    faces = np.column_stack((v, v + 1, v + nx + 2, v + nx + 1))
    return xu.Ugrid2d(node_x, node_y, -1, faces)


def test_polygonize__errors(grid):
    uda = xu.UgridDataArray(
        xr.DataArray(np.ones(grid.n_edge), dims=[grid.edge_dimension]), grid=grid
    )
    with pytest.raises(ValueError, match="Cannot polygonize non-face dimension"):
        xu.polygonize(uda)

    uda = xu.UgridDataArray(
        xr.DataArray(np.ones((3, grid.n_face)), dims=["layer", grid.face_dimension]),
        grid=grid,
    )
    with pytest.raises(ValueError, match="Cannot polygonize non-face dimension"):
        xu.polygonize(uda)


def test_polygonize(grid):
    a = np.array([0, 0, 0, 1, 1, 1, 0, 0, 0])
    uda = xu.UgridDataArray(xr.DataArray(a, dims=grid.face_dimension), grid)
    actual = xu.polygonize(uda)
    assert isinstance(actual, gpd.GeoDataFrame)
    assert len(actual) == 3

    # With a hole in the 1-valued polygon.
    a = np.array([1, 1, 1, 1, 0, 1, 1, 1, 1])
    uda = xu.UgridDataArray(xr.DataArray(a, dims=grid.face_dimension), grid)
    actual = xu.polygonize(uda)
    assert isinstance(actual, gpd.GeoDataFrame)
    assert len(actual) == 2



import numpy as np
import pytest
from scipy import sparse

from xugrid.constants import FILL_VALUE
from xugrid.ugrid import connectivity


@pytest.fixture(scope="function")
def triangle_mesh():
    # Two triangles
    faces = np.array(
        [
            [0, 1, 2],
            [1, 3, 2],
        ]
    )
    return faces


@pytest.fixture(scope="function")
def mixed_mesh():
    # Triangle, quadrangle
    faces = np.array(
        [
            [0, 1, 2, FILL_VALUE],
            [1, 3, 4, 2],
        ]
    )
    return faces


def test_argsort_rows():
    with pytest.raises(ValueError, match="Array is not 2D"):
        connectivity.argsort_rows(np.array([3, 2, 1, 0]))

    array = np.array(
        [
            [1, 0],
            [0, 1],
            [2, 2],
            [2, 1],
            [0, 2],
            [2, 0],
        ]
    )
    _, expected = np.unique(array, axis=0, return_index=True)
    actual = connectivity.argsort_rows(array)
    assert np.array_equal(actual, expected)


def test_index_like():
    xy_a = np.array([[0.0, 0.0], [1.0, 1.0]])
    xy_b = np.array([[0.0, 0.0]])
    with pytest.raises(ValueError, match="coordinates do not match in shape"):
        connectivity.index_like(xy_a, xy_b, tolerance=0.0)

    xy_b = np.array([[0.0, 0.0], [1.1, 1.0]])
    with pytest.raises(ValueError, match="coordinates are not identical after sorting"):
        connectivity.index_like(xy_a, xy_b, tolerance=0.0)
    # Now with higher tolerance
    connectivity.index_like(xy_a, xy_b, tolerance=0.2)

    xy_a = np.array([[3.0, 3.0], [1.0, 1.0], [2.0, 2.0], [0.0, 0.0]])
    xy_b = np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0], [2.0, 2.0]])
    actual = connectivity.index_like(xy_a, xy_b, tolerance=0.0)
    expected = [3, 1, 0, 2]
    assert np.array_equal(actual, expected)

    # With tolerance
    xy_a = np.array([[3.0, 3.0001], [1.0, 1.0], [2.0, 2.0], [-0.0001, 0.0]])
    xy_b = np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0], [2.0, 2.0]])
    actual = connectivity.index_like(xy_a, xy_b, tolerance=0.001)
    expected = [3, 1, 0, 2]
    assert np.array_equal(actual, expected)


def test_neighbors():
    i = [0, 0, 0, 1, 1, 1]
    j = [0, 1, 2, 1, 3, 2]
    coo_content = (j, (i, j))
    A = sparse.coo_matrix(coo_content).tocsr()
    n, m = A.shape
    A = connectivity.AdjacencyMatrix(A.indices, A.indptr, A.nnz, n, m)
    assert np.array_equal(connectivity.neighbors(A, 0), [0, 1, 2])
    assert np.array_equal(connectivity.neighbors(A, 1), [1, 2, 3])


def test_to_ij(triangle_mesh, mixed_mesh):
    faces = triangle_mesh
    actual_i, actual_j = connectivity._to_ij(faces, invert=False)
    expected_i = [0, 0, 0, 1, 1, 1]
    expected_j = [0, 1, 2, 1, 3, 2]
    assert np.array_equal(actual_i, expected_i)
    assert np.array_equal(actual_j, expected_j)

    # Inverted
    actual_i, actual_j = connectivity._to_ij(faces, invert=True)
    assert np.array_equal(actual_i, expected_j)
    assert np.array_equal(actual_j, expected_i)

    faces = mixed_mesh
    actual_i, actual_j = connectivity._to_ij(faces, invert=False)
    expected_i = [0, 0, 0, 1, 1, 1, 1]
    expected_j = [0, 1, 2, 1, 3, 4, 2]
    assert np.array_equal(actual_i, expected_i)
    assert np.array_equal(actual_j, expected_j)

    # Inverted
    actual_i, actual_j = connectivity._to_ij(faces, invert=True)
    assert np.array_equal(actual_i, expected_j)
    assert np.array_equal(actual_j, expected_i)


def test_to_sparse(mixed_mesh):
    faces = mixed_mesh
    csr = connectivity._to_sparse(faces, invert=False, sort_indices=True)
    expected_j = np.array([0, 1, 2, 1, 2, 3, 4])
    assert np.array_equal(csr.indices, expected_j)
    assert csr.has_sorted_indices

    csr = connectivity._to_sparse(faces, invert=False, sort_indices=False)
    expected_j = np.array([0, 1, 2, 1, 3, 4, 2])
    assert np.array_equal(csr.indices, expected_j)
    assert not csr.has_sorted_indices


def test_ragged_index():
    n = 3
    m = 4
    m_per_row = np.array([1, 2, 3])
    actual = connectivity.ragged_index(n, m, m_per_row)
    expected = np.array(
        [
            [True, False, False, False],
            [True, True, False, False],
            [True, True, True, False],
        ]
    )
    assert np.array_equal(actual, expected)


def test_sparse_dense_conversion_roundtrip(triangle_mesh, mixed_mesh):
    faces = triangle_mesh
    sparse = connectivity.to_sparse(faces)
    back = connectivity.to_dense(sparse)
    # Note: roundtrip does not preserve CW/CCW orientation, since orientation
    # does not apply to node_face_connectivity, but the sorted rows should
    # contain the same elements.
    assert np.array_equal(faces.sort(axis=1), back.sort(axis=1))

    faces = mixed_mesh
    sparse = connectivity.to_sparse(faces)
    back = connectivity.to_dense(sparse)
    assert np.array_equal(faces.sort(axis=1), back.sort(axis=1))


def test_invert_dense(triangle_mesh, mixed_mesh):
    faces = triangle_mesh
    actual = connectivity.invert_dense(faces)
    expected = np.array(
        [
            [0, -1],  # 0
            [0, 1],  # 1
            [0, 1],  # 2
            [1, -1],  # 3
        ]
    )
    assert np.array_equal(actual, expected)

    faces = mixed_mesh
    actual = connectivity.invert_dense(faces)
    expected = np.array(
        [
            [0, -1],  # 0
            [0, 1],  # 1
            [0, 1],  # 2
            [1, -1],  # 3
            [1, -1],  # 4
        ]
    )
    assert np.array_equal(actual, expected)


def test_invert_sparse(triangle_mesh, mixed_mesh):
    faces = triangle_mesh
    sparse = connectivity.to_sparse(faces)
    inverted = connectivity.invert_sparse(sparse)
    actual = connectivity.to_dense(inverted)
    expected = np.array(
        [
            [0, -1],  # 0
            [0, 1],  # 1
            [0, 1],  # 2
            [1, -1],  # 3
        ]
    )
    assert np.array_equal(actual, expected)

    faces = mixed_mesh
    sparse = connectivity.to_sparse(faces)
    inverted = connectivity.invert_sparse(sparse)
    actual = connectivity.to_dense(inverted)
    expected = np.array(
        [
            [0, -1],  # 0
            [0, 1],  # 1
            [0, 1],  # 2
            [1, -1],  # 3
            [1, -1],  # 4
        ]
    )
    assert np.array_equal(actual, expected)


def test_to_dense(triangle_mesh):
    faces = triangle_mesh
    sparse = connectivity.to_sparse(faces)
    actual = connectivity.to_dense(sparse)
    assert np.array_equal(actual, np.sort(faces, axis=1))

    with pytest.raises(ValueError, match="n_columns 2 is too small"):
        connectivity.to_dense(sparse, n_columns=2)

    # now pad
    actual = connectivity.to_dense(sparse, n_columns=4)
    expected = np.array(
        [
            [0, 1, 2, FILL_VALUE],
            [1, 2, 3, FILL_VALUE],
        ]
    )
    assert np.array_equal(actual, expected)

    # and twice
    actual = connectivity.to_dense(sparse, n_columns=5)
    expected = np.array(
        [
            [0, 1, 2, FILL_VALUE, FILL_VALUE],
            [1, 2, 3, FILL_VALUE, FILL_VALUE],
        ]
    )
    assert np.array_equal(actual, expected)


def test_renumber():
    a = np.array(
        [
            [0, 1, 2],
            [10, 11, 12],
            [30, 31, 32],
        ]
    )
    actual = connectivity.renumber(a)
    expected = np.array(
        [
            [0, 1, 2],
            [3, 4, 5],
            [6, 7, 8],
        ]
    )
    assert np.array_equal(actual, expected)

    a = np.array(
        [
            [0, 1, 2],
            [10, 11, 2],
            [30, 31, 2],
        ]
    )
    actual = connectivity.renumber(a)
    expected = np.array(
        [
            [0, 1, 2],
            [3, 4, 2],
            [5, 6, 2],
        ]
    )
    assert np.array_equal(actual, expected)


def test_renumber_with_fill_value():
    a = np.array(
        [
            [0, 1, FILL_VALUE],
            [10, 11, 12],
            [30, FILL_VALUE, 32],
        ]
    )
    actual = connectivity.renumber(a)
    expected = np.array(
        [
            [0, 1, -1],
            [2, 3, 4],
            [5, -1, 6],
        ]
    )
    assert np.array_equal(actual, expected)

    a = np.array(
        [
            [0, -1, 2],
            [10, 11, 2],
            [30, -1, 2],
        ]
    )
    actual = connectivity.renumber(a)
    expected = np.array(
        [
            [0, -1, 1],
            [2, 3, 1],
            [4, -1, 1],
        ]
    )
    assert np.array_equal(actual, expected)


def test_close_polygons(mixed_mesh):
    faces = mixed_mesh
    closed, isfill = connectivity.close_polygons(faces)
    expected = np.array(
        [
            [0, 1, 2, 0, 0],
            [1, 3, 4, 2, 1],
        ]
    )
    expected_isfill = np.full((2, 5), False)
    expected_isfill[0, -2:] = True
    expected_isfill[1, -1] = True
    assert np.array_equal(closed, expected)
    assert np.array_equal(isfill, expected_isfill)


def test_reverse_orientation(mixed_mesh):
    faces = mixed_mesh
    reverse = connectivity.reverse_orientation(faces)
    expected = np.array(
        [
            [2, 1, 0, FILL_VALUE],
            [2, 4, 3, 1],
        ]
    )
    assert np.array_equal(reverse, expected)


def test_counterclockwise():
    nodes = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [2.0, 0.0],
            [0.0, 2.0],
        ]
    )

    # Already ccw, nothing should be changed.
    faces = np.array([[0, 2, 3, -1]])
    actual = connectivity.counterclockwise(faces, nodes)
    assert np.array_equal(actual, faces)

    # Clockwise with a fill value, reverse.
    faces_cw = np.array([[3, 2, 0, -1]])
    actual = connectivity.counterclockwise(faces_cw, nodes)
    assert np.array_equal(actual, faces)

    # Including a hanging node, ccw, nothing changed.
    hanging_ccw = np.array([[0, 1, 2, 3, -1]])
    actual = connectivity.counterclockwise(hanging_ccw, nodes)
    assert np.array_equal(actual, hanging_ccw)

    # Including a hanging node, reverse.
    hanging_cw = np.array([[3, 2, 1, 0, -1]])
    actual = connectivity.counterclockwise(hanging_cw, nodes)
    assert np.array_equal(actual, hanging_ccw)


def test_edge_connectivity(mixed_mesh):
    faces = mixed_mesh
    edge_nodes, face_edges = connectivity.edge_connectivity(faces)
    expected_edge_nodes = np.array(
        [
            [0, 1],
            [0, 2],
            [1, 2],
            [1, 3],
            [2, 4],
            [3, 4],
        ]
    )
    expected_face_edges = np.array(
        [
            [0, 2, 1, -1],
            [3, 5, 4, 2],
        ]
    )
    assert np.array_equal(edge_nodes, expected_edge_nodes)
    assert np.array_equal(face_edges, expected_face_edges)


def test_validate_edge_connectivity(mixed_mesh):
    faces = mixed_mesh
    edges = np.array([[0, 1]])
    with pytest.raises(ValueError, match="face_node_connectivity defines 6 edges"):
        connectivity.validate_edge_node_connectivity(faces, edges)

    edges = np.array(
        [
            [0, 1],  # T
            [0, 1],  # F
            [1, 0],  # F
            [0, 2],  # T
            [1, 2],  # T
            [1, 3],  # T
            [2, 4],  # T
            [3, 4],  # T
            [0, 4],  # F
        ]
    )
    actual = connectivity.validate_edge_node_connectivity(faces, edges)
    expected = np.array([True, False, False, True, True, True, True, True, False])
    assert np.array_equal(actual, expected)


def test_node_node_connectivity():
    edge_nodes = np.array(
        [
            [0, 1],  # 0
            [1, 2],  # 1
            [2, 3],  # 2
            [3, 0],  # 3
            [1, 4],  # 4
            [4, 5],  # 5
            [5, 2],  # 6
        ]
    )
    csr = connectivity.node_node_connectivity(edge_nodes)
    coo = csr.tocoo()
    actual = np.column_stack([coo.row, coo.col])
    expected = np.array(
        [
            [0, 1],
            [0, 3],
            [1, 0],
            [1, 2],
            [1, 4],
            [2, 1],
            [2, 3],
            [2, 5],
            [3, 0],
            [3, 2],
            [4, 1],
            [4, 5],
            [5, 2],
            [5, 4],
        ]
    )
    expected_edges = np.array([0, 3, 0, 1, 4, 1, 2, 6, 3, 2, 4, 5, 6, 5])
    assert np.array_equal(actual, expected)
    assert np.array_equal(expected_edges, coo.data)


def test_directed_node_node_connectivity():
    edge_nodes = np.array(
        [
            [0, 1],  # 0
            [1, 2],  # 1
            [2, 3],  # 2
            [3, 0],  # 3
            [1, 4],  # 4
            [4, 5],  # 5
            [5, 2],  # 6
            [0, 6],  # 7
        ]
    )
    csr = connectivity.directed_node_node_connectivity(edge_nodes)
    coo = csr.tocoo()
    actual = np.column_stack([coo.row, coo.col])
    expected = np.array(
        [
            [0, 1],
            [0, 6],
            [1, 2],
            [1, 4],
            [2, 3],
            [3, 0],
            [4, 5],
            [5, 2],
        ]
    )
    expected_edges = np.array([0, 7, 1, 4, 2, 3, 5, 6])
    assert np.array_equal(actual, expected)
    assert np.array_equal(expected_edges, coo.data)


class TestEdgeConnectivity:
    r"""
    Test networks:
    
    Simple:

                    3
                   /
                  / 2
       0     1   /
    0 --- 1 --- 2 
                 \
                  \ 3
                   \
                    4 

    And:

                             5
                          4 --- 6
                         /
                        / 3
       0     1     2   /
    0 --- 1 --- 2 --- 3 
                       \
                        \ 4
                         \
                          5 
    """

    @pytest.fixture(autouse=True)
    def setup(self):
        self.simple_edge_node_connectivity = np.array(
            [
                [0, 1],
                [1, 2],
                [2, 3],
                [2, 4],
            ]
        )
        self.simple_node_edge_connectivity = connectivity.invert_dense_to_sparse(
            self.simple_edge_node_connectivity
        )
        self.edge_node_connectivity = np.array(
            [
                [0, 1],  # edge 0
                [1, 2],  # edge 1
                [2, 3],  # edge 2
                [3, 4],  # edge 3
                [3, 5],  # edge 4
                [4, 6],  # edge 5
            ]
        )
        self.node_edge_connectivity = connectivity.invert_dense_to_sparse(
            self.edge_node_connectivity
        )

    def test_edge_edge_connectivity_simple(self):
        csr = connectivity.edge_edge_connectivity(
            self.simple_edge_node_connectivity, self.simple_node_edge_connectivity
        )
        assert isinstance(csr, sparse.csr_matrix)
        coo = csr.tocoo()
        actual = np.column_stack([coo.row, coo.col])
        expected = np.array(
            [[0, 1], [1, 0], [1, 2], [1, 3], [2, 1], [2, 3], [3, 1], [3, 2]]
        )
        assert np.array_equal(actual, expected)
        # Test through which node the connection is formed.
        assert np.array_equal(coo.data, [1, 1, 2, 2, 2, 2, 2, 2])

    def test_direct_edge_edge_connectivity_simple(self):
        csr = connectivity.directed_edge_edge_connectivity(
            self.simple_edge_node_connectivity, self.simple_node_edge_connectivity
        )
        coo = csr.tocoo()
        actual = np.column_stack([coo.row, coo.col])
        expected = np.array([[0, 1], [1, 2], [1, 3]])
        assert np.array_equal(actual, expected)
        # Test through which node the connection is formed.
        assert np.array_equal(coo.data, [1, 2, 2])

    def test_directed_edge_edge_connectivity(self):
        csr = connectivity.directed_edge_edge_connectivity(
            self.edge_node_connectivity, self.node_edge_connectivity
        )
        assert isinstance(csr, sparse.csr_matrix)

        coo = csr.tocoo()
        actual = np.column_stack([coo.row, coo.col])
        expected = np.array([[0, 1], [1, 2], [2, 3], [2, 4], [3, 5]])
        assert np.array_equal(actual, expected)
        # Test through which node the connection is formed.
        assert np.array_equal(coo.data, [1, 2, 3, 3, 4])


def test_face_face_connectivity():
    edge_faces = np.array(
        [
            [0, -1],
            [0, -1],
            [0, 1],
            [1, -1],
            [1, -1],
            [1, -1],
        ]
    )
    face_face = connectivity.face_face_connectivity(edge_faces, n_face=2)
    assert isinstance(face_face, sparse.csr_matrix)
    assert np.array_equal(face_face.indices, [1, 0])
    assert np.array_equal(face_face.indptr, [0, 1, 2])
    assert np.array_equal(face_face.data, [2, 2])

    # Test two isolated faces
    # Two triangles:
    edge_faces = np.array(
        [
            [0, -1],
            [0, -1],
            [0, -1],
            [1, -1],
            [1, -1],
            [1, -1],
        ]
    )
    face_face = connectivity.face_face_connectivity(edge_faces, n_face=2)
    assert isinstance(face_face, sparse.csr_matrix)
    assert face_face.shape == (2, 2)


def test_centroids(mixed_mesh):
    faces = mixed_mesh
    nodes = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [2.0, 0.0],
            [2.0, 1.0],
        ]
    )
    actual = connectivity.centroids(faces, nodes[:, 0], nodes[:, 1])
    expected = np.array(
        [
            [2.0 / 3.0, 1.0 / 3.0],
            [1.5, 0.5],
        ]
    )
    assert np.allclose(actual, expected)


def test_circumcenters_error(mixed_mesh):
    faces = mixed_mesh
    nodes = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [2.0, 0.0],
            [2.0, 1.0],
        ]
    )
    with pytest.raises(NotImplementedError):
        connectivity.circumcenters(faces, nodes[:, 0], nodes[:, 1])


def test_circumcenters(triangle_mesh):
    faces = triangle_mesh
    nodes = np.array(
        [
            [0.0, 1.0],
            [0.0, 0.0],
            [2.0, 0.0],
            [1.0, -2.0],
        ]
    )
    actual = connectivity.circumcenters(faces, nodes[:, 0], nodes[:, 1])
    expected = np.array([[1.0, 0.5], [1.0, -0.75]])
    assert np.allclose(actual, expected)


def test_structured_connectivity():
    active = np.array(
        [
            [True, True, False],
            [True, True, True],
            [True, False, True],
        ]
    )
    A = connectivity.structured_connectivity(active)
    assert A.nnz == 14
    assert np.array_equal(connectivity.neighbors(A, 0), [1, 2])
    assert np.array_equal(connectivity.neighbors(A, 1), [0, 3])
    assert np.array_equal(connectivity.neighbors(A, 2), [0, 3, 5])
    assert np.array_equal(connectivity.neighbors(A, 3), [1, 2, 4])
    assert np.array_equal(connectivity.neighbors(A, 4), [3, 6])
    assert np.array_equal(connectivity.neighbors(A, 5), [2])
    assert np.array_equal(connectivity.neighbors(A, 6), [4])


def test_area(mixed_mesh):
    node_x = np.array([0.0, 1.0, 1.0, 2.0, 2.0])
    node_y = np.array([0.0, 0.0, 1.0, 0.0, 1.0])
    actual = connectivity.area(mixed_mesh, node_x, node_y)
    assert np.allclose(actual, [0.5, 1.0])


def test_perimeter(mixed_mesh):
    node_x = np.array([0.0, 1.0, 1.0, 2.0, 2.0])
    node_y = np.array([0.0, 0.0, 1.0, 0.0, 1.0])
    actual = connectivity.perimeter(mixed_mesh, node_x, node_y)
    assert np.allclose(actual, [2.0 + np.sqrt(2), 4.0])


def test_triangulate(mixed_mesh):
    faces = mixed_mesh
    actual_triangles, actual_faces = connectivity.triangulate_dense(faces)
    expected_triangles = np.array(
        [
            [0, 1, 2],
            [1, 3, 4],
            [1, 4, 2],
        ]
    )
    expected_faces = np.array([0, 1, 1])
    assert np.array_equal(actual_triangles, expected_triangles)
    assert np.array_equal(actual_faces, expected_faces)

    sparse_faces = connectivity.to_sparse(faces, sort_indices=False).tocoo()
    actual_triangles, actual_faces = connectivity.triangulate_coo(sparse_faces)
    assert np.array_equal(actual_triangles, expected_triangles)
    assert np.array_equal(actual_faces, expected_faces)


def test_binary_erosion():
    i = np.array([0, 1, 1, 2, 2, 3, 3])
    j = np.array([1, 0, 2, 1, 3, 2, 4])
    coo_content = (j, (i, j))
    con = sparse.coo_matrix(coo_content).tocsr()
    a = np.full(5, True)

    with pytest.raises(ValueError, match="Binary operations"):
        connectivity.binary_erosion(con, a[np.newaxis, :])

    actual = connectivity.binary_erosion(con, a)
    assert actual.all()

    exterior = np.array([0, 4])
    actual = connectivity.binary_erosion(con, a, exterior=exterior)
    expected = np.array([False, True, True, True, False])
    assert np.array_equal(actual, expected)
    # Check for mutation
    assert a.all()

    actual = connectivity.binary_erosion(con, a, exterior=exterior, iterations=3)
    assert (~actual).all()

    mask = np.array([False, False, False, True, True])
    actual = connectivity.binary_erosion(
        con, a, exterior=exterior, iterations=3, mask=mask
    )
    assert np.array_equal(actual, mask)

    a = np.array([False, True, True, True, False])
    actual = connectivity.binary_erosion(con, a)
    expected = np.array([False, False, True, False, False])
    assert np.array_equal(actual, expected)


def test_binary_dilation():
    i = np.array([0, 1, 1, 2, 2, 3, 3])
    j = np.array([1, 0, 2, 1, 3, 2, 4])
    coo_content = (j, (i, j))
    con = sparse.coo_matrix(coo_content).tocsr()
    a = np.full(5, False)

    with pytest.raises(ValueError, match="Binary operations"):
        connectivity.binary_dilation(con, a[np.newaxis, :])

    # No change
    actual = connectivity.binary_dilation(con, a)
    assert (~actual).all()

    exterior = np.array([0, 4])
    actual = connectivity.binary_dilation(con, a, exterior=exterior)
    assert (~actual).all()

    actual = connectivity.binary_dilation(con, a, exterior=exterior, border_value=True)
    expected = np.array([True, False, False, False, True])
    assert np.array_equal(actual, expected)
    # Check for mutation
    assert (~a).all()

    actual = connectivity.binary_dilation(
        con, a, exterior=exterior, iterations=3, border_value=True
    )
    assert actual.all()

    mask = np.array([False, False, False, True, True])
    actual = connectivity.binary_dilation(
        con, a, exterior=exterior, iterations=3, mask=mask, border_value=True
    )
    assert np.array_equal(actual, ~mask)

    a = np.array([False, False, True, False, False])
    actual = connectivity.binary_dilation(con, a)
    expected = np.array([False, True, True, True, False])
    assert np.array_equal(actual, expected)


@pytest.fixture(scope="function")
def dag() -> sparse.csr_matrix:
    #
    #    0──►2──►3
    #    │   ▲
    #    │   │
    #    └──►1
    #
    i = [0, 1, 0, 2]
    j = [1, 2, 2, 3]
    csr = sparse.coo_matrix((j, (i, j)), shape=(4, 4)).tocsr()
    return csr


@pytest.fixture(scope="function")
def cycle() -> sparse.csr_matrix:
    # same as dag fixture, but bidirectional
    i = [0, 1, 0, 2]
    j = [1, 2, 2, 3]
    ij = np.concatenate((i, j))
    ji = np.concatenate((j, i))
    csr = sparse.coo_matrix((ji, (ij, ji)), shape=(4, 4)).tocsr()
    return csr


def test_topological_sort_by_dfs(dag):
    actual = connectivity.topological_sort_by_dfs(dag)
    assert np.array_equal(actual, [0, 1, 2, 3])


def test_topological_sort_by_dfs__cycle_error(cycle):
    with pytest.raises(ValueError, match="The graph contains at least one cycle"):
        connectivity.topological_sort_by_dfs(cycle)


def test_contract_vertices(dag):
    actual = connectivity.contract_vertices(dag, [0, 1, 3])
    expected = np.array([[0, 1], [1, 3]])
    assert np.array_equal(actual, expected)


def test_contract_vertices__cycle_error(cycle):
    with pytest.raises(ValueError, match="The graph contains at least one cycle"):
        connectivity.contract_vertices(cycle, indices=[1, 3])



import geopandas as gpd
import numpy as np
import pytest
import shapely
import xarray as xr

from xugrid import conversion as cv
from xugrid.constants import FILL_VALUE


@pytest.fixture(scope="function")
def line():
    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    edge_node_connectivity = np.array(
        [
            [0, 1],
            [1, 2],
        ]
    )
    return x, y, edge_node_connectivity


@pytest.fixture(scope="function")
def line_gdf():
    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    gdf = gpd.GeoDataFrame(geometry=[shapely.linestrings(x, y)])
    return gdf


@pytest.fixture(scope="function")
def triangle_mesh():
    x = np.array([0.0, 1.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 1.0, 0.0])
    # Two triangles
    faces = np.array(
        [
            [0, 1, 2],
            [1, 3, 2],
        ]
    )
    return x, y, faces


@pytest.fixture(scope="function")
def mixed_mesh():
    x = np.array([0.0, 1.0, 1.0, 2.0, 2.0])
    y = np.array([0.0, 0.0, 1.0, 0.0, 1.0])
    # Triangle, quadrangle
    faces = np.array(
        [
            [0, 1, 2, FILL_VALUE],
            [1, 3, 4, 2],
        ]
    )
    return x, y, faces


@pytest.fixture(scope="function")
def structured_mesh_ascending():
    da = xr.DataArray(
        data=np.arange(12).reshape((3, 4)),
        coords={"y": [5.0, 10.0, 15.0], "x": [2.0, 4.0, 6.0, 8.0]},
        dims=["y", "x"],
        name="grid",
    )
    return da


@pytest.fixture(scope="function")
def structured_mesh_descending():
    da = xr.DataArray(
        data=np.arange(12).reshape((3, 4)),
        coords={"y": [15.0, 10.0, 5.0], "x": [8.0, 6.0, 4.0, 2.0]},
        dims=["y", "x"],
        name="grid",
    )
    return da


def test_nodes_geos_roundtrip(line):
    x, y, _ = line
    actual = cv.nodes_to_points(x, y)
    x_back, y_back = cv.points_to_nodes(actual)
    points_back = cv.nodes_to_points(x_back, y_back)
    assert np.array_equal(x, x_back)
    assert np.array_equal(y, y_back)
    assert np.array_equal(actual, points_back)


def test_linestrings_to_edges(line_gdf):
    x, y, segments = cv.linestrings_to_edges(line_gdf.geometry.values)
    assert np.allclose(x, [0.0, 1.0, 2.0])
    assert np.allclose(y, [0.0, 0.0, 0.0])
    assert np.array_equal(segments, [[0, 1], [1, 2]])


def test_edges_geos_roundtrip(line):
    x, y, c = line
    actual = cv.edges_to_linestrings(x, y, c)
    x_back, y_back, c_back = cv.linestrings_to_edges(actual)
    lines_back = cv.edges_to_linestrings(x_back, y_back, c_back)
    assert np.array_equal(x, x_back)
    assert np.array_equal(y, y_back)
    assert np.array_equal(c, c_back)
    assert np.array_equal(actual, lines_back)


# Cannot use fixtures in parametrize:
# https://github.com/pytest-dev/pytest/issues/349
def _faces_geos_roundtrip(mesh):
    x, y, c = mesh
    actual = cv.faces_to_polygons(x, y, c)
    x_back, y_back, c_back = cv.polygons_to_faces(actual)
    polygons_back = cv.faces_to_polygons(x_back, y_back, c_back)
    assert np.array_equal(x, x_back)
    assert np.array_equal(y, y_back)
    assert np.array_equal(c, c_back)
    assert np.array_equal(actual, polygons_back)


def test_faces_geos_roundtrip__triangle(triangle_mesh):
    _faces_geos_roundtrip(triangle_mesh)


def test_faces_geos_roundtrip__mixed(mixed_mesh):
    _faces_geos_roundtrip(mixed_mesh)


def test_is_monotonic_and_increasing():
    with pytest.raises(ValueError):
        cv._is_monotonic_and_increasing([0.0, -1.0, 2.0])
    with pytest.raises(ValueError):
        cv._is_monotonic_and_increasing([2.0, 0.0, 1.0])

    assert cv._is_monotonic_and_increasing([0.0, 1.0, 2.0])
    assert not cv._is_monotonic_and_increasing([2.0, 1.0, 0.0])

    ascending = np.array(
        [
            [0.0, 1.0, 2.0],
            [3.0, 4.0, 5.0],
            [6.0, 7.0, 8.0],
        ]
    )
    descending = np.array(
        [
            [8.0, 7.0, 6.0],
            [5.0, 4.0, 3.0],
            [2.0, 1.0, 0.0],
        ]
    )
    assert cv._is_monotonic_and_increasing(ascending, axis=0)
    assert cv._is_monotonic_and_increasing(ascending, axis=1)
    assert not cv._is_monotonic_and_increasing(descending, axis=1)
    assert not cv._is_monotonic_and_increasing(descending, axis=1)


def test_infer_interval_breaks():
    assert np.allclose([-0.5, 0.5, 1.5], cv.infer_interval_breaks([0, 1]))
    assert np.allclose(
        [-0.5, 0.5, 5.0, 9.5, 10.5], cv.infer_interval_breaks([0, 1, 9, 10])
    )

    xref, yref = np.meshgrid(np.arange(6), np.arange(5))
    cx = (xref[1:, 1:] + xref[:-1, :-1]) / 2
    cy = (yref[1:, 1:] + yref[:-1, :-1]) / 2
    x = cv.infer_interval_breaks(cx, axis=1)
    x = cv.infer_interval_breaks(x, axis=0)
    y = cv.infer_interval_breaks(cy, axis=1)
    y = cv.infer_interval_breaks(y, axis=0)
    np.testing.assert_allclose(xref, x)
    np.testing.assert_allclose(yref, y)

    # test that ValueError is raised for non-monotonic 1D inputs
    with pytest.raises(ValueError):
        cv.infer_interval_breaks(np.array([0, 2, 1]), check_monotonic=True)


def test_scalar_spacing(structured_mesh_ascending, structured_mesh_descending):
    upcoords = structured_mesh_ascending.coords["x"]
    downcoords = structured_mesh_descending.coords["x"]
    spacing = xr.DataArray(0.1, name="dx")
    with pytest.raises(ValueError, match="spacing of x does not match value of dx"):
        cv._scalar_spacing(upcoords, spacing)

    spacing = xr.DataArray(2.0, name="dx")
    assert np.allclose(cv._scalar_spacing(upcoords, spacing), 1.0)
    spacing = xr.DataArray(-2.0, name="dx")
    assert np.allclose(cv._scalar_spacing(downcoords, spacing), 1.0)


def test_array_spacing(structured_mesh_ascending, structured_mesh_descending):
    upcoords = structured_mesh_ascending.coords["x"]
    downcoords = structured_mesh_descending.coords["x"]
    spacing = xr.DataArray([0.1], name="dx")
    with pytest.raises(ValueError, match="size of x does not match size of dx"):
        cv._array_spacing(upcoords, spacing)

    spacing = xr.DataArray([2.0, 2.0, 2.0, 2.0], name="dx")
    assert np.allclose(cv._array_spacing(upcoords, spacing), 1.0)
    spacing = xr.DataArray([-2.0, -2.0, -2.0, -2.0], name="dx")
    assert np.allclose(cv._array_spacing(downcoords, spacing), 1.0)


def test_implicit_spacing(structured_mesh_ascending, structured_mesh_descending):
    da = xr.DataArray(
        [[0.0, 0.0]],
        {"y": [0.0], "x": [1.0, 2.0]},
        ["y", "x"],
    )
    with pytest.raises(ValueError, match="Cannot derive spacing of 1-sized coordinate"):
        cv.infer_interval_breaks1d(da, "y")

    actual = cv.infer_interval_breaks1d(structured_mesh_ascending, "x")
    assert np.allclose(actual, [1.0, 3.0, 5.0, 7.0, 9.0])
    actual = cv.infer_interval_breaks1d(structured_mesh_descending, "x")
    assert np.allclose(actual, [9.0, 7.0, 5.0, 3.0, 1.0])


@pytest.mark.parametrize("spacing_type", ["implicit", "scalar", "array"])
def test_infer_breaks_intervals1d(
    structured_mesh_ascending, structured_mesh_descending, spacing_type
):
    up = structured_mesh_ascending
    down = structured_mesh_descending
    x_expected = np.array(
        [1.0, 3.0, 5.0, 7.0, 9.0],
    )
    y_expected = np.array(
        [2.5, 7.5, 12.5, 17.5],
    )

    if spacing_type == "scalar":
        up = up.assign_coords({"dx": 2.0, "dy": 5.0})
        down = down.assign_coords({"dx": 2.0, "dy": 5.0})
    elif spacing_type == "array":
        up = up.assign_coords({"dx": ("x", [2.0] * 4), "dy": ("y", [5.0] * 3)})
        down = down.assign_coords({"dx": ("x", [2.0] * 4), "dy": ("y", [5.0] * 3)})

    assert np.allclose(cv.infer_interval_breaks1d(up, "x"), x_expected)
    assert np.allclose(cv.infer_interval_breaks1d(up, "y"), y_expected)
    assert np.allclose(cv.infer_interval_breaks1d(down, "x"), x_expected[::-1])
    assert np.allclose(cv.infer_interval_breaks1d(down, "y"), y_expected[::-1])


def test_infer_breaks_intervals1d_errors(structured_mesh_ascending):
    up = structured_mesh_ascending
    up = up.assign_coords(x=[2.0, 4.0, 3.0, 8.0])

    with pytest.raises(ValueError, match="The input coordinate is not monotonic."):
        cv.infer_interval_breaks1d(up, "x")


def test_bounds1d_to_vertices():
    with pytest.raises(ValueError, match="Bounds are not monotonic"):
        cv.bounds1d_to_vertices(
            xr.DataArray(
                data=[[0.0, 1.0], [2.0, 3.0], [1.0, 2.0]], dims=["x", "nbound"]
            )
        )

    x_vertices = np.array([1.0, 3.0, 5.0, 7.0, 9.0])
    y_vertices = np.array([2.5, 7.5, 12.5, 17.5])
    # Ascending
    x_bounds = np.column_stack((x_vertices[:-1], x_vertices[1:]))
    y_bounds = np.column_stack((y_vertices[:-1], y_vertices[1:]))
    assert np.allclose(cv.bounds1d_to_vertices(x_bounds), x_vertices)
    assert np.allclose(cv.bounds1d_to_vertices(y_bounds), y_vertices)
    # Descending
    xrev = x_vertices[::-1]
    yrev = y_vertices[::-1]
    x_bounds = np.column_stack((xrev[1:], xrev[:-1]))
    y_bounds = np.column_stack((yrev[1:], yrev[:-1]))
    assert np.allclose(cv.bounds1d_to_vertices(x_bounds), xrev)
    assert np.allclose(cv.bounds1d_to_vertices(y_bounds), yrev)


def test_bounds2d_to_topology2d():
    # Clockwise
    x_bounds = np.array(
        [[[0.0, 0.0, 1.0, 1.0], [2.0, 2.0, 3.0, 3.0], [4.0, 4.0, 5.0, 5.0]]]
    )
    y_bounds = np.array(
        [[[0.0, 1.0, 1.0, 0.0], [2.0, 3.0, 3.0, 2.0], [4.0, 5.0, 5.0, 4.0]]]
    )
    x, y, faces, index = cv.bounds2d_to_topology2d(x_bounds, y_bounds)

    assert index.all()
    assert faces.shape == (3, 4)

    # Result should be counterclockwise
    expected_first_cell = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )
    actual = np.column_stack((x, y))[faces[0]]
    assert np.allclose(actual, expected_first_cell)


def test_bounds2d_to_topology2d_invalid_coordinate():
    # Test with some invalid (NaN) coordinates
    x_bounds = np.array(
        [[[0.0, 0.0, 1.0, 1.0], [2.0, 2.0, 3.0, 3.0], [4.0, 4.0, 5.0, 5.0]]]
    )
    y_bounds = np.array(
        [[[0.0, 1.0, 1.0, 0.0], [2.0, 3.0, 3.0, 2.0], [4.0, 5.0, 5.0, 4.0]]]
    )
    x_bounds_with_nan = x_bounds.copy()
    x_bounds_with_nan[0, 0, 0] = np.nan

    with pytest.warns(
        UserWarning,
        match="A UGRID2D face requires at least three unique non-collinear vertices.",
    ):
        _, _, faces, index = cv.bounds2d_to_topology2d(x_bounds_with_nan, y_bounds)

    # Test that the first cell is marked invalid
    assert not index[0]
    assert index[1:].all()
    assert faces.shape == (2, 4)


def test_bounds2d_to_topology2d_bad_bounds():
    # Test bad bounds. Triangles are allowed, points and lines are not.
    x_bounds = np.array(
        [
            [
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 2.0, 2.0, 1.0],
                [2.0, 3.0, 3.0, 2.0],
                [2.0, 2.0, 3.0, 3.0],
            ]
        ]
    )
    y_bounds = np.array(
        [
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 1.0, 1.0],
            ]
        ]
    )

    with pytest.warns(
        UserWarning,
        match="A UGRID2D face requires at least three unique non-collinear vertices.",
    ):
        x, _, faces, index = cv.bounds2d_to_topology2d(x_bounds, y_bounds)

    assert np.array_equal(index, [False, True, True, False])
    expected_x = [
        [1.0, 2.0, 2.0, 1.0],
        [2.0, 3.0, 2.0, 3.0],  # repeated node moved to last one
    ]
    assert np.array_equal(x[faces], expected_x)


def test_bounds2d_to_topology2d_collinear_vertices():
    x_bounds = np.array(
        [[[0.0, 0.33, 0.67, 1.0], [2.0, 2.0, 3.0, 3.0], [4.0, 4.0, 5.0, 5.0]]]
    )
    y_bounds = np.array(
        [[[0.5, 0.5, 0.5, 0.5], [2.0, 3.0, 3.0, 2.0], [4.0, 5.0, 5.0, 4.0]]]
    )
    with pytest.warns(
        UserWarning,
        match="A UGRID2D face requires at least three unique non-collinear vertices.",
    ):
        _, _, faces, index = cv.bounds2d_to_topology2d(x_bounds, y_bounds)
        assert len(faces) == 2
        assert np.array_equal(index, [False, True, True])


def test_infer_xy_coords():
    da = xr.DataArray(
        data=[[1]],
        coords={"y": [1], "x": [1]},
        dims=["y", "x"],
    )
    assert cv.infer_xy_coords(da) == ("x", "y")

    da = xr.DataArray(
        data=[[1]],
        coords={"latitude": [1], "longitude": [1]},
        dims=["latitude", "longitude"],
    )
    assert cv.infer_xy_coords(da) == ("longitude", "latitude")

    da = xr.DataArray(
        data=[[1]],
        coords={"lat": [1], "lon": [1]},
        dims=["lat", "lon"],
    )
    assert cv.infer_xy_coords(da) == (None, None)

    da["lon"].attrs["axis"] = "X"
    da["lat"].attrs["axis"] = "Y"

    da = xr.DataArray(
        data=[[1]],
        coords={"lat": [1], "lon": [1]},
        dims=["lat", "lon"],
    )
    da["lon"].attrs["standard_name"] = "longitude"
    da["lat"].attrs["standard_name"] = "latitude"
    assert cv.infer_xy_coords(da) == ("lon", "lat")

    da = xr.DataArray(
        data=[[1]],
        coords={"yy": [1], "xx": [1]},
        dims=["yy", "xx"],
    )
    da["xx"].attrs["standard_name"] = "projection_x_coordinate"
    da["yy"].attrs["standard_name"] = "projection_y_coordinate"
    assert cv.infer_xy_coords(da) == ("xx", "yy")



from xugrid import data
from xugrid.core.common import (
    concat,
    full_like,
    load_dataarray,
    load_dataset,
    merge,
    ones_like,
    open_dataarray,
    open_dataset,
    open_mfdataset,
    open_zarr,
    zeros_like,
)
from xugrid.core.dataarray_accessor import UgridDataArrayAccessor
from xugrid.core.dataset_accessor import UgridDatasetAccessor
from xugrid.core.wrap import UgridDataArray, UgridDataset
from xugrid.plot import plot
from xugrid.regrid.gridder import NetworkGridder
from xugrid.regrid.regridder import (
    BarycentricInterpolator,
    CentroidLocatorRegridder,
    OverlapRegridder,
    RelativeOverlapRegridder,
)
from xugrid.ugrid.burn import burn_vector_geometry, earcut_triangulate_polygons
from xugrid.ugrid.conventions import UgridRolesAccessor
from xugrid.ugrid.partitioning import merge_partitions
from xugrid.ugrid.polygonize import polygonize
from xugrid.ugrid.snapping import (
    create_snap_to_grid_dataframe,
    snap_nodes,
    snap_to_grid,
)
from xugrid.ugrid.ugrid1d import Ugrid1d
from xugrid.ugrid.ugrid2d import Ugrid2d

__version__ = "0.15.1"

__all__ = (
    "data",
    "concat",
    "full_like",
    "load_dataarray",
    "load_dataset",
    "merge",
    "ones_like",
    "open_dataarray",
    "open_dataset",
    "open_mfdataset",
    "open_zarr",
    "zeros_like",
    "UgridDataArrayAccessor",
    "UgridDatasetAccessor",
    "UgridDataArray",
    "UgridDataset",
    "plot",
    "BarycentricInterpolator",
    "CentroidLocatorRegridder",
    "OverlapRegridder",
    "RelativeOverlapRegridder",
    "burn_vector_geometry",
    "earcut_triangulate_polygons",
    "NetworkGridder",
    "UgridRolesAccessor",
    "merge_partitions",
    "polygonize",
    "snap_nodes",
    "snap_to_grid",
    "create_snap_to_grid_dataframe",
    "Ugrid1d",
    "Ugrid2d",
)



from itertools import chain
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr
from numba_celltree import EdgeCellTree2d
from numpy.typing import ArrayLike
from scipy import sparse

import xugrid
from xugrid import conversion
from xugrid.constants import (
    FILL_VALUE,
    BoolArray,
    FloatArray,
    FloatDType,
    IntArray,
    IntDType,
    LineArray,
)
from xugrid.core.utils import either_dict_or_kwargs
from xugrid.regrid.utils import alt_cumsum
from xugrid.ugrid import connectivity, conventions
from xugrid.ugrid.selection_utils import section_coordinates_1d
from xugrid.ugrid.ugridbase import AbstractUgrid, as_pandas_index


class Ugrid1d(AbstractUgrid):
    """
    This class stores the topological data of a "1-D unstructured grid": a
    collection of connected line elements, such as a river network.

    Parameters
    ----------
    node_x: ndarray of floats
    node_y: ndarray of floats
    fill_value: int
    edge_node_connectivity: ndarray of integers
    name: string, optional
        Network name. Defaults to "network1d".
    dataset: xr.Dataset, optional
    indexes: Dict[str, str], optional
        When a dataset is provided, a mapping from the UGRID role to the dataset
        variable name. E.g. {"face_x": "mesh2d_face_lon"}.
    is_projected: bool, optional
        Whether node_x and node_y are longitude and latitude or projected x and
        y coordinates. Used to write the appropriate standard_name in the
        coordinate attributes. If crs is provided, its value will take priority.
    crs: Any, optional
        Coordinate Reference System of the geometry objects. Can be anything accepted by
        :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
        such as an authority string (eg "EPSG:4326") or a WKT string.
    attrs: Dict[str, str], optional
        UGRID topology attributes. Should not be provided together with
        dataset: if other names are required, update the dataset instead.
        A name entry is ignored, as name is given explicitly.
    start_index: int, 0 or 1, default is 0.
        Start index of the connectivity arrays. Must match the start index
        of the provided face_node_connectivity and edge_node_connectivity.
    """

    def __init__(
        self,
        node_x: FloatArray,
        node_y: FloatArray,
        fill_value: int,
        edge_node_connectivity: IntArray = None,
        name: str = "network1d",
        dataset: xr.Dataset = None,
        indexes: Dict[str, str] = None,
        is_projected: bool = True,
        crs: Any = None,
        attrs: Dict[str, str] = None,
        start_index: int = 0,
    ):
        self.node_x = np.ascontiguousarray(node_x)
        self.node_y = np.ascontiguousarray(node_y)
        self.fill_value = fill_value
        self.start_index = start_index
        self.edge_node_connectivity = edge_node_connectivity - self.start_index
        self.name = name

        # projected, crs
        self.crs, self.is_projected = self._validate_crs(crs, is_projected)

        self._initialize_indexes_attrs(name, dataset, indexes, attrs)
        self._dataset = dataset

        # Optional attributes, deferred initialization
        # Meshkernel
        self._mesh = None
        self._meshkernel = None
        # Celltree
        self._celltree = None
        self._node_kdtree = None
        self._edge_kdtree = None
        # Bounds
        self._xmin = None
        self._xmax = None
        self._ymin = None
        self._ymax = None
        # Edges
        self._edge_x = None
        self._edge_y = None
        # Connectivity
        self._node_node_connectivity = None
        self._node_edge_connectivity = None

    @classmethod
    def from_dataset(cls, dataset: xr.Dataset, topology: str = None):
        """
        Extract the 1D UGRID topology information from an xarray Dataset.

        Parameters
        ----------
        dataset: xr.Dataset
            Dataset containing topology information stored according to UGRID conventions.

        Returns
        -------
        grid: Ugrid1dAdapter
        """
        ds = dataset
        if not isinstance(ds, xr.Dataset):
            raise TypeError(
                "Ugrid should be initialized with xarray.Dataset. "
                f"Received instead: {type(ds)}"
            )
        if topology is None:
            topology = cls._single_topology(ds)

        indexes = {}

        # Collect names
        connectivity = ds.ugrid_roles.connectivity[topology]
        coordinates = ds.ugrid_roles.coordinates[topology]
        ugrid_vars = (
            [topology]
            + list(connectivity.values())
            + list(chain.from_iterable(chain.from_iterable(coordinates.values())))
        )

        # Take the first coordinates by default.
        # They can be reset with .set_node_coords()
        x_index = coordinates["node_coordinates"][0][0]
        y_index = coordinates["node_coordinates"][1][0]
        node_x_coordinates = ds[x_index].astype(FloatDType).to_numpy()
        node_y_coordinates = ds[y_index].astype(FloatDType).to_numpy()

        edge_nodes = connectivity["edge_node_connectivity"]
        fill_value = ds[edge_nodes].encoding.get("_FillValue", -1)
        start_index = ds[edge_nodes].attrs.get("start_index", 0)
        edge_node_connectivity = cls._prepare_connectivity(
            ds[edge_nodes], fill_value, dtype=IntDType
        ).to_numpy()

        # Fill "indexes": mark which names point to the UGRID-relevant coordinates.
        indexes["node_x"] = x_index
        indexes["node_y"] = y_index
        edge_indexes = coordinates.get("edge_coordinates")
        if edge_indexes is not None:
            indexes["edge_x"] = edge_indexes[0][0]
            indexes["edge_y"] = edge_indexes[1][0]

        crs, is_projected = cls._extract_crs(ds, topology)

        return cls(
            node_x_coordinates,
            node_y_coordinates,
            fill_value,
            edge_node_connectivity,
            name=topology,
            dataset=dataset[ugrid_vars],
            indexes=indexes,
            is_projected=is_projected,
            crs=crs,
            start_index=start_index,
        )

    def _clear_geometry_properties(self):
        """Clear all properties that may have been invalidated"""
        # Meshkernel
        self._mesh = None
        self._meshkernel = None
        # Celltree
        self._celltree = None
        self._node_kdtree = None
        self._edge_kdtree = None
        # Bounds
        self._xmin = None
        self._xmax = None
        self._ymin = None
        self._ymax = None
        # Edges
        self._edge_x = None
        self._edge_y = None

    def _assign_derived_coords(self, obj):
        if self.node_dimension in obj.dims:
            obj = self.assign_node_coords(obj)
        if self.edge_dimension in obj.dims:
            obj = self.assign_edge_coords(obj)
        if self._dataset is not None:
            self._dataset = self._dataset.drop_vars(
                self._indexes.values(), errors="ignore"
            )
        return obj

    @classmethod
    def from_meshkernel(
        cls,
        mesh,
        name: str = "network1d",
        is_projected: bool = True,
        crs: Any = None,
    ):
        """
        Create a 1D UGRID topology from a MeshKernel Mesh1d object.

        Parameters
        ----------
        mesh: MeshKernel.Mesh2d
        name: str
            Mesh name. Defaults to "network1d".
        is_projected: bool
            Whether node_x and node_y are longitude and latitude or projected x and
            y coordinates. Used to write the appropriate standard_name in the
            coordinate attributes.
        crs: Any, optional
            Coordinate Reference System of the geometry objects. Can be anything accepted by
            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.

        Returns
        -------
        grid: Ugrid1d
        """
        return cls(
            mesh.node_x,
            mesh.node_y,
            fill_value=FILL_VALUE,
            edge_node_connectivity=mesh.edge_nodes.reshape((-1, 2)),
            name=name,
            is_projected=is_projected,
            crs=crs,
        )

    def to_dataset(
        self, other: xr.Dataset = None, optional_attributes: bool = False
    ) -> xr.Dataset:
        node_x = self._indexes["node_x"]
        node_y = self._indexes["node_y"]
        edge_nodes = self._attrs["edge_node_connectivity"]
        edge_nodes_attrs = conventions.DEFAULT_ATTRS["edge_node_connectivity"]

        data_vars = {
            self.name: 0,
            edge_nodes: xr.DataArray(
                data=self._adjust_connectivity(self.edge_node_connectivity),
                attrs=edge_nodes_attrs,
                dims=(self.edge_dimension, "two"),
            ),
        }

        attrs = {"Conventions": "CF-1.9 UGRID-1.0"}
        if other is not None:
            attrs.update(other.attrs)

        dataset = xr.Dataset(data_vars, attrs=attrs)
        if self._dataset:
            dataset = dataset.merge(self._dataset, compat="override")
        if other is not None:
            dataset = dataset.merge(other)
        if node_x not in dataset or node_y not in dataset:
            dataset = self.assign_node_coords(dataset)
        if optional_attributes:
            dataset = self.assign_edge_coords(dataset)

        dataset[self.name].attrs = self._filtered_attrs(dataset)
        dataset = self.write_grid_mapping(dataset)
        return dataset

    @property
    def topology_dimension(self):
        """Highest dimensionality of the geometric elements: 1"""
        return 1

    @property
    def core_dimension(self):
        return self.edge_dimension

    @property
    def dims(self):
        """Set of UGRID dimension names: node dimension, edge dimension."""
        return {self.node_dimension, self.edge_dimension}

    @property
    def sizes(self):
        return {self.node_dimension: self.n_node, self.edge_dimension: self.n_edge}

    @property
    def coords(self):
        """Dictionary for grid coordinates."""
        return {
            self.node_dimension: self.node_coordinates,
            self.edge_dimension: self.edge_coordinates,
        }

    def get_coordinates(self, dim: str) -> FloatArray:
        """Return the coordinates for the specified UGRID dimension."""
        if dim == self.node_dimension:
            return self.node_coordinates
        elif dim == self.edge_dimension:
            return self.edge_coordinates
        else:
            raise ValueError(
                f"Expected {self.node_dimension} or {self.edge_dimension}; got: {dim}"
            )

    @property
    def facets(self) -> dict[str, str]:
        return {
            "node": self.node_dimension,
            "edge": self.edge_dimension,
        }

    def get_connectivity_matrix(self, dim: str, xy_weights: bool):
        """Return the connectivity matrix for the specified UGRID dimension."""
        if dim == self.node_dimension:
            connectivity = self.node_node_connectivity.copy()
            coordinates = self.node_coordinates
        else:
            raise ValueError(f"Expected {self.node_dimension}; got: {dim}")

        if xy_weights:
            connectivity.data = self._connectivity_weights(connectivity, coordinates)

        return connectivity

    def _locate_nearest(
        self, facet: str, points: FloatArray, max_distance: float = np.inf
    ):
        if facet == "node":
            return self.locate_nearest_node(points, max_distance)
        elif facet == "edge":
            return self.locate_nearest_edge(points, max_distance)
        else:
            raise ValueError(
                f"Expected facet as one of {'node', 'edge'}, received: {facet}"
            )

    # These are all optional attributes. They are not computed by default, only
    # when called upon.

    @property
    def mesh(self) -> "mk.Mesh1d":  # type: ignore # noqa
        """
        Create if needed, and return meshkernel Mesh1d object.

        Returns
        -------
        mesh: meshkernel.Mesh1d
        """
        import meshkernel as mk

        if self._mesh is None:
            edge_nodes = self.edge_node_connectivity.ravel().astype(np.int32)
            self._mesh = mk.Mesh1d(
                node_x=self.node_x,
                node_y=self.node_y,
                edge_nodes=edge_nodes,
            )
        return self._mesh

    @property
    def meshkernel(self) -> "mk.MeshKernel":  # type: ignore # noqa
        """
        Create if needed, and return meshkernel MeshKernel instance.

        Returns
        -------
        meshkernel: meshkernel.MeshKernel
        """
        import meshkernel as mk

        if self._meshkernel is None:
            if self.is_geographic:
                mk_projection = mk.ProjectionType.SPHERICAL
            else:
                mk_projection = mk.ProjectionType.CARTESIAN
            self._meshkernel = mk.MeshKernel(mk_projection)
            self._meshkernel.mesh1d_set(self.mesh)
        return self._meshkernel

    @property
    def is_cyclic(self) -> bool:
        """
        Return True if the directed node node connectivity contains cycles. If
        False, then the directed node node connectivity is a directed acyclic
        graph (DAG).

        Runs a depth-first-search.
        """
        # topological sort returns a ValueError if it detects a cycle.
        try:
            self.topological_sort_by_dfs()
            return False
        except ValueError as e:
            if str(e) == "The graph contains at least one cycle":
                return True
            raise  # Re-raise any other ValueError

    @classmethod
    def from_geodataframe(cls, geodataframe: "geopandas.GeoDataFrame") -> "Ugrid1d":  # type: ignore # noqa
        """
        Convert geodataframe of linestrings into a UGRID1D topology.

        Parameters
        ----------
        geodataframe: geopandas GeoDataFrame

        Returns
        -------
        topology: Ugrid1d
        """
        import geopandas as gpd

        if not isinstance(geodataframe, gpd.GeoDataFrame):
            raise TypeError(
                f"Expected GeoDataFrame, received: {type(geodataframe).__name__}"
            )
        return cls.from_shapely(geodataframe.geometry.to_numpy(), crs=geodataframe.crs)

    @staticmethod
    def from_shapely(geometry: LineArray, crs=None) -> "Ugrid1d":
        """
        Convert an array of shapely linestrings to UGRID1D topology.

        Parameters
        ----------
        geometry: np.ndarray of shapely linestrings
        crs: Any, optional
            Coordinate Reference System of the geometry objects. Can be anything accepted by
            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        """

        import shapely

        if not (shapely.get_type_id(geometry) == shapely.GeometryType.LINESTRING).all():
            raise TypeError(
                "Can only create Ugrid1d from shapely LineString geometries, "
                "geometry contains other types of geometries."
            )

        x, y, edge_node_connectivity = conversion.linestrings_to_edges(geometry)
        return Ugrid1d(x, y, FILL_VALUE, edge_node_connectivity, crs=crs)

    def to_pygeos(self, dim):
        from warnings import warn

        warn(
            ".to_pygeos has been deprecated. Use .to_shapely instead.",
            DeprecationWarning,
        )
        return self.to_shapely(dim)

    def to_shapely(self, dim):
        """
        Convert UGRID topology to shapely objects.

        * nodes: points
        * edges: linestrings

        Parameters
        ----------
        dim: str
            Node or edge dimension.

        Returns
        -------
        geometry: ndarray of shapely.Geometry
        """
        if dim == self.node_dimension:
            return conversion.nodes_to_points(
                self.node_x,
                self.node_y,
            )
        elif dim == self.edge_dimension:
            return conversion.edges_to_linestrings(
                self.node_x,
                self.node_y,
                self.edge_node_connectivity,
            )
        else:
            raise ValueError(
                f"Dimension {dim} is not a node or edge dimension of the"
                " Ugrid1d topology."
            )

    def isel(self, indexers=None, return_index=False, **indexers_kwargs):
        """
        Select based on node or edge.

        Edge selection always results in a valid UGRID topology. Node selection
        may result in invalid topologies (incomplete edges), and will error in
        such a case.

        Parameters
        ----------
        indexers: dict of str to np.ndarray of integers or bools
        return_index: bool, optional
            Whether to return node_index, edge_index.

        Returns
        -------
        obj: xr.Dataset or xr.DataArray
        grid: Ugrid2d
        indexes: dict
            Dictionary with keys node dimension, edge dimension and values
            their respective index. Only returned if return_index is True.
        """
        indexers = either_dict_or_kwargs(indexers, indexers_kwargs, "isel")
        alldims = self.dims
        invalid = indexers.keys() - alldims
        if invalid:
            raise ValueError(
                f"Dimensions {invalid} do not exist. Expected one of {alldims}"
            )

        indexers = {k: as_pandas_index(v, self.sizes[k]) for k, v in indexers.items()}
        nodedim = self.node_dimension
        edgedim = self.edge_dimension

        edge_index = {}
        if nodedim in indexers:
            node_index = indexers[nodedim]
            edge_index[nodedim] = np.unique(
                self.node_edge_connectivity[node_index].data
            )
        if edgedim in indexers:
            edge_index[edgedim] = indexers[edgedim]

        # Convert all to pandas index.
        edge_index = {k: as_pandas_index(v, self.n_edge) for k, v in edge_index.items()}

        # Check the indexes against each other.
        index = self._precheck(edge_index)
        grid, finalized_indexers = self.topology_subset(index, return_index=True)
        self._postcheck(indexers, finalized_indexers)

        if return_index:
            return grid, finalized_indexers
        else:
            return grid

    def _validate_indexer(self, indexer) -> Tuple[float, float]:
        if isinstance(indexer, slice):
            if indexer.step is not None:
                raise ValueError("Ugrid1d does not support steps in slices")
            if indexer.start >= indexer.stop:
                raise ValueError("slice start should be smaller than slice stop")
        else:
            raise ValueError("Ugrid1d only supports slice indexing")
        return indexer.start, indexer.stop

    def sel(self, obj, x, y) -> Tuple:
        """
        Select a selection of edges, based on edge centroids.

        Parameters
        ----------
        x: slice
        y: slice

        Returns
        -------
        dimension: str
        as_ugrid: bool
        index: 1d array of integers
        coords: dict
        """
        xmin, xmax = self._validate_indexer(x)
        ymin, ymax = self._validate_indexer(y)
        edge_index = np.nonzero(
            (self.edge_x >= xmin)
            & (self.edge_x < xmax)
            & (self.edge_y >= ymin)
            & (self.edge_y < ymax)
        )[0]
        grid, indexes = self.topology_subset(edge_index, return_index=True)
        indexes = {k: v for k, v in indexes.items() if k in obj.dims}
        new_obj = obj.isel(indexes)
        return new_obj, grid

    def topology_subset(
        self, edge_index: Union[BoolArray, IntArray], return_index: bool = False
    ):
        """
        Create a new UGRID1D topology for a subset of this topology.

        Parameters
        ----------
        edge_index: 1d array of integers or bool
            Edges of the subset.
        return_index: bool, optional
            Whether to return node_index, edge_index.

        Returns
        -------
        subset: Ugrid1d
        indexes: dict
            Dictionary with keys node dimension and edge dimension and values
            their respective index. Only returned if return_index is True.
        """
        if not isinstance(edge_index, pd.Index):
            edge_index = as_pandas_index(edge_index, self.n_edge)

        range_index = pd.RangeIndex(0, self.n_edge)
        if edge_index.size == self.n_edge and edge_index.equals(range_index):
            if return_index:
                indexes = {
                    self.node_dimension: pd.RangeIndex(0, self.n_node),
                    self.edge_dimension: range_index,
                }
                return self, indexes
            else:
                return self

        # N.B. edges do not contain fill values, as there are always two nodes
        # required to form an edge.
        edge_subset = self.edge_node_connectivity[edge_index]
        node_index = np.unique(edge_subset.ravel())
        new_edges = connectivity.renumber(edge_subset)
        node_x = self.node_x[node_index]
        node_y = self.node_y[node_index]
        grid = Ugrid1d(
            node_x,
            node_y,
            FILL_VALUE,
            new_edges,
            name=self.name,
            indexes=self._indexes,
            is_projected=self.is_projected,
            crs=self.crs,
            attrs=self._attrs,
        )
        self._propagate_properties(grid)
        if return_index:
            indexes = {
                self.node_dimension: pd.Index(node_index),
                self.edge_dimension: edge_index,
            }
            return grid, indexes
        else:
            return grid

    def clip_box(
        self,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
    ):
        return self.sel(x=slice(xmin, xmax), y=slice(ymin, ymax))

    @property
    def celltree(self) -> EdgeCellTree2d:
        """
        Initializes the celltree if needed, and returns celltree.

        A celltree is a search structure for spatial lookups in unstructured grids.
        """
        if self._celltree is None:
            self._celltree = EdgeCellTree2d(
                self.node_coordinates,
                self.edge_node_connectivity,
            )
        return self._celltree

    @staticmethod
    def _section_coordinates(
        edges: FloatArray, xy: FloatArray, dim: str, index: IntArray, name: str
    ):
        return section_coordinates_1d(edges, xy, dim, index, name)

    def to_periodic(self, obj):
        return self, obj

    def to_nonperiodic(self, xmax, obj):
        return self, obj

    def topological_sort_by_dfs(self) -> IntArray:
        """
        Return an array of vertices in topological order.

        Returns
        -------
        sorted_vertices: np.ndarray of integer
        """
        return connectivity.topological_sort_by_dfs(
            self.directed_node_node_connectivity
        )

    def remove_self_loops(self) -> "Ugrid1d":
        """
        Remove degenerate edges: those that join a node to itself, also called
        a self-loop. These edges have a length of exactly zero.

        Returns
        -------
        grid: Ugrid1d
        """
        a, b = self.edge_node_connectivity.T
        not_self_loop = a != b
        edge_subset = self.edge_node_connectivity[not_self_loop]
        valid = np.bincount(edge_subset.ravel()) > 0
        new_edges = connectivity.renumber(edge_subset)
        return Ugrid1d(
            node_x=self.node_x[valid],
            node_y=self.node_y[valid],
            fill_value=self.fill_value,
            edge_node_connectivity=new_edges,
            name=self.name,
            indexes=self._indexes,
            is_projected=self.is_projected,
            crs=self.crs,
            attrs=self._attrs,
        )

    def contract_vertices(self, indices: IntArray) -> "Ugrid1d":
        """
        Return a simplified network topology by removing all nodes that are
        not listed in ``indices``.

        Parameters
        ----------
        indices: np.ndarray of integers

        Returns
        -------
        contracted: Ugrid1d
        """
        edges = connectivity.contract_vertices(
            self.directed_node_node_connectivity, indices
        )
        node_index = np.unique(edges.ravel())
        new_edges = connectivity.renumber(edges)
        return Ugrid1d(
            node_x=self.node_x[node_index],
            node_y=self.node_y[node_index],
            fill_value=self.fill_value,
            edge_node_connectivity=new_edges,
            name=self.name,
            indexes=self._indexes,
            is_projected=self.is_projected,
            crs=self.crs,
            attrs=self._attrs,
        )

    def refine_by_vertices(
        self,
        vertices: FloatArray,
        return_index: bool = False,
        tolerance: Optional[float] = None,
    ) -> "Ugrid1d":
        """
        Refine Ugrid1d with extra vertices to be inserted and returns new grid.
        Vertices need to be located on existing grid edges, if not, a ValueError
        will be returned.

        Parameters
        ----------
        vertices: np.ndarray of floats
            Coordinates of vertices to be inserted in the grid. Must have shape
            (N, 2).
        return_index: bool, optional
            If set to to True, the index of the new vertices in the grid will be
            returned. Defaults to False.
        tolerance: float, optional
            The tolerance used to determine whether a point is on an edge. This
            accounts for the inherent inexactness of floating point calculations.
            If None, an appropriate tolerance is automatically estimated based on
            the geometry size. Consider adjusting this value if edge detection
            results are unsatisfactory.

        Returns
        -------
        grid: Ugrid1d
            Refined grid with new vertices.

        Examples
        --------
        Let's first create a simple grid with 3 nodes and 2 edges:

        >>> import numpy as np
        >>> import xugrid as xu
        >>> node_xy = np.array([[0.0, 0.0], [5.0, 5.0], [10.0, 5.0]])
        >>> edge_nodes = np.array([[0, 1],[1, 2]])
        >>> grid = xu.Ugrid1d(*node_xy.T, -1, edge_nodes)

        Now refine the grid by adding new vertices:

        >>> vertices = np.array([[2.0, 2.0], [7.0, 5.0]])
        >>> new = grid.refine_by_vertices(vertices)
        >>> print(new.node_coordinates)

        To return the indices of the inserted vertices:

        >>> new, new_vertices_index = grid.refine_by_vertices(vertices, return_index=True)
        >>> print(new_vertices_index)
        >>> print(new.node_coordinates[new_vertices_index])

        """
        edge_index = self.celltree.locate_points(vertices, tolerance)
        invalid = edge_index == -1
        if invalid.any():
            raise ValueError(
                f"The following vertices are not located on any edge:\n{vertices[invalid]}"
            )

        # Do not insert vertices that are already present in the grid.
        node_xy = self.node_coordinates
        combined = np.concatenate((node_xy, vertices))
        _, index, inverse = np.unique(
            combined, return_index=True, return_inverse=True, axis=0
        )
        index_to_vertices = index[inverse][self.n_node :]
        not_duplicated = index_to_vertices >= self.n_node
        new_vertices = vertices[not_duplicated]
        edge_index = edge_index[not_duplicated]

        first_node = self.edge_node_connectivity[edge_index, 0]
        distance = np.linalg.norm(new_vertices - node_xy[first_node], axis=1)
        grid_edge_index = np.arange(self.n_edge)
        repeats = np.bincount(np.concatenate((grid_edge_index, edge_index)))
        new_edges = np.repeat(self.edge_node_connectivity, repeats, axis=0)
        order = np.lexsort((distance, edge_index))

        # Index for new vertices
        node_index = np.arange(self.n_node, self.n_node + len(edge_index))[order]

        # For the new edges, modify:
        #
        # * all but the last entry per edge of the second column
        # * all but the first entry per edge of the first column
        #
        i = np.arange(len(new_edges))
        mask0 = np.repeat(alt_cumsum(repeats), repeats)
        mask1 = np.repeat(np.cumsum(repeats), repeats) - 1
        new_edges[i > mask0, 0] = node_index
        new_edges[i < mask1, 1] = node_index

        grid = Ugrid1d(
            np.concatenate((self.node_x, new_vertices[:, 0])),
            np.concatenate((self.node_y, new_vertices[:, 1])),
            self.fill_value,
            new_edges,
            name=self.name,
            is_projected=self.is_projected,
            crs=self.crs,
        )
        self._propagate_properties(grid)
        if return_index:
            return grid, node_index
        else:
            return grid

    def _nearest_interpolate(
        self,
        data: FloatArray,
        ugrid_dim: str,
        max_distance: float,
    ) -> FloatArray:
        isnull = np.isnan(data)
        if isnull.all():
            raise ValueError("All values are NA.")

        edge_length = self.edge_length
        if ugrid_dim == self.node_dimension:
            # Set the edge length as the graph weights.
            # We can do this easily since the data of the node_node_connectivity CSR matrix
            # contains the edge through which the connection is made.
            connectivity = self.node_node_connectivity.copy()
            connectivity.data = edge_length[connectivity.data]
        elif ugrid_dim == self.edge_dimension:
            # Convert to COO-form so we can index with rows and columns.
            connectivity = self.edge_edge_connectivity.tocoo()
            # Compute distance from edge centroid to centroid along the edges.
            # I.e. half of both edge lengths
            connectivity.data = 0.5 * (
                edge_length[connectivity.row] + edge_length[connectivity.col]
            )
        else:
            raise ValueError(
                f"Expected {self.node_dimension} or {self.edge_dimension}, "
                f"received instead: {ugrid_dim}"
            )
        _, _, index = sparse.csgraph.dijkstra(
            csgraph=connectivity,
            indices=np.flatnonzero(~isnull),
            return_predecessors=True,
            limit=max_distance,
            min_only=True,
        )
        # dijkstra returns -9999 when no path could be found.
        # There, we'll keep the NaN value.
        found = index != -9999
        index = index[found]
        out = data.copy()
        out[found] = data[index]
        return out

    @staticmethod
    def merge_partitions(
        grids: Sequence["Ugrid1d"],
    ) -> tuple["Ugrid1d", dict[str, np.array]]:
        """
        Merge grid partitions into a single whole.

        Duplicate edges are included only once, and removed from subsequent
        partitions before merging.

        Parameters
        ----------
        grids: sequence of Ugrid1d

        Returns
        -------
        merged: Ugrid1d
        """
        from xugrid.ugrid import partitioning

        # Grab a sample grid
        grid = next(iter(grids))
        fill_value = grid.fill_value
        node_coordinates, node_indexes, node_inverse = partitioning.merge_nodes(grids)
        new_edges, edge_indexes = partitioning.merge_edges(grids, node_inverse)
        indexes = {
            grid.node_dimension: node_indexes,
            grid.edge_dimension: edge_indexes,
        }

        merged_grid = Ugrid1d(
            *node_coordinates.T,
            fill_value,
            new_edges,
            name=grid.name,
            indexes=grid._indexes,
            is_projected=grid.is_projected,
            crs=grid.crs,
            attrs=grid._attrs,
        )
        return merged_grid, indexes

    def reindex_like(
        self,
        other: "Ugrid1d",
        obj: Union[xr.DataArray, xr.Dataset],
        tolerance: float = 0.0,
    ):
        """
        Conform a DataArray or Dataset to match the topology of another Ugrid1D
        topology. The topologies must be exactly equivalent: only the order of
        the nodes and edges may differ.

        Parameters
        ----------
        other: Ugrid1d
        obj: DataArray or Dataset
        tolerance: float, default value 0.0.
            Maximum distance between inexact coordinate matches.

        Returns
        -------
        reindexed: DataArray or Dataset
        """
        if not isinstance(other, Ugrid1d):
            raise TypeError(f"Expected Ugrid1d, received: {type(other).__name__}")

        indexers = {
            self.node_dimension: connectivity.index_like(
                xy_a=self.node_coordinates,
                xy_b=other.node_coordinates,
                tolerance=tolerance,
            ),
            self.edge_dimension: connectivity.index_like(
                xy_a=self.edge_coordinates,
                xy_b=other.edge_coordinates,
                tolerance=tolerance,
            ),
        }
        return obj.isel(indexers, missing_dims="ignore")

    def create_data_array(self, data: ArrayLike, facet: str) -> "xugrid.UgridDataArray":
        """
        Create a UgridDataArray from this grid and a 1D array of values.

        Parameters
        ----------
        data: array like
            Values for this array. Must be a ``numpy.ndarray`` or castable to
            it.
        grid: Ugrid1d, Ugrid2d
        facet: str
            With which facet to associate the data. Options for Ugrid1d are,
            ``"node"`` or ``"edge"``. Options for Ugrid2d are ``"node"``,
            ``"edge"``, or ``"face"``.

        Returns
        -------
        uda: UgridDataArray
        """
        if facet == "node":
            dimension = self.node_dimension
        elif facet == "edge":
            dimension = self.edge_dimension
        else:
            raise ValueError(f"Invalid facet: {facet}. Must be one of: node, edge.")
        return self._create_data_array(data, dimension)



from typing import Tuple

import numpy as np

from xugrid.constants import FloatArray, IntArray


def get_sorted_section_coords(
    s: FloatArray, xy: FloatArray, dim: str, index: IntArray, name: str
):
    order = np.argsort(s)
    coords = {
        f"{name}_x": (dim, xy[order, 0]),
        f"{name}_y": (dim, xy[order, 1]),
        f"{name}_s": (dim, s[order]),
    }
    return coords, index[order]


def section_coordinates_1d(
    edges: FloatArray, xy: FloatArray, dim: str, index: IntArray, name: str
) -> Tuple[IntArray, dict]:
    s = np.linalg.norm(xy - edges[0, 0], axis=1)
    return get_sorted_section_coords(s, xy, dim, index, name)


def section_coordinates_2d(
    edges: FloatArray, xy: FloatArray, dim: str, index: IntArray, name: str
) -> Tuple[IntArray, dict]:
    # TODO: add boundaries xy[:, 0] and xy[:, 1]
    xy_mid = 0.5 * (xy[:, 0, :] + xy[:, 1, :])
    return section_coordinates_1d(edges, xy_mid, dim, index, name)



from typing import Tuple

import numpy as np
from scipy import sparse

from xugrid.constants import FILL_VALUE, IntArray


def _bbox_area(bounds):
    return (bounds[2] - bounds[0]) * (bounds[3] - bounds[1])


def _classify(
    i: IntArray, j: IntArray, face_values: np.ndarray
) -> Tuple[int, IntArray]:
    """
    Find out how many discrete polygons are created. Identify the connectivity,
    such that we can select a single polygon afterwards.

    Parameters
    ----------
    i: np.ndarray of int
        First face of the edge.
    j: np.ndarray of int
        Second face of the edge.
    face_values: np.ndarray

    Returns
    -------
    n_polygon: int
    polygon_id: np.ndarray of int
    """
    # Every face connects up to two faces. vi holds the values of the first face,
    # vj holds the value of the second face.
    vi = face_values[i]
    vj = face_values[j]
    n = face_values.size
    # For labelling, only those parts of the mesh that have the same value
    # should be connected with each other.
    # Since we dropped NaN values before, we needn't worry about those.
    is_connection = (i != FILL_VALUE) & (j != FILL_VALUE) & (vi == vj)
    i = i[is_connection]
    j = j[is_connection]
    ij = np.concatenate([i, j])
    ji = np.concatenate([j, i])
    coo_content = (ji, (ij, ji))
    # Make sure to explicitly set the matrix shape: otherwise, isolated
    # elements witout any connection might disappear, and connected_components
    # will not return a value for every face.
    coo_matrix = sparse.coo_matrix(coo_content, shape=(n, n))
    # We can classify the grid faces using this (reduced) connectivity
    return sparse.csgraph.connected_components(coo_matrix)


def polygonize(uda: "UgridDataArray") -> "gpd.GeoDataFrame":  # type: ignore # noqa
    """
    Polygonize a UgridDataArray.

    This function creates vector polygons for all connected regions of cells
    (faces) in the Ugrid2d topology sharing a common value.

    The produced polygon edges will follow exactly the cell boundaries. When
    the data consists of many unique values (e.g. unbinned elevation data), the
    result will essentially be one polygon per face. In such cases, it is much
    more efficient to use ``xugrid.UgridDataArray.to_geodataframe``, which
    directly converts every cell to a polygon. This function is meant for data
    with relatively few unique values such as classification results.

    Parameters
    ----------
    uda: UgridDataArray
        The DataArray should only contain the face dimension. Additional
        dimensions, such as time, are not allowed.

    Returns
    -------
    polygonized: GeoDataFrame
    """

    import geopandas as gpd
    import shapely

    facedim = uda.ugrid.grid.face_dimension
    if uda.dims != (facedim,):
        raise ValueError(
            "Cannot polygonize non-face dimensions. Expected only"
            f"({facedim},), but received {uda.dims}."
        )

    # First remove the NaN values. These will not be polygonized anyway.
    dropped = uda.dropna(dim=uda.ugrid.grid.face_dimension)
    face_values = dropped.to_numpy()
    grid = dropped.ugrid.grid
    i, j = grid.edge_face_connectivity.T
    n_polygon, polygon_id = _classify(i, j, face_values)

    # Now we identify for each label the subset of edges. These are the
    # "exterior" edges: either the exterior edge of the mesh identified by a
    # fill value, or by being connected to a cell with a different value.
    coordinates = grid.node_coordinates
    data_i = face_values[i]
    vi = polygon_id[i]
    vj = polygon_id[j]
    # Ensure that no result thas has been created by indexing with the
    # fill_value remains. Since polygon_id starts counting a 0, we may use -1.
    vi[i == FILL_VALUE] = FILL_VALUE
    vj[j == FILL_VALUE] = FILL_VALUE
    boundary = vi != vj

    polygons = []
    values = []
    for label in range(n_polygon):
        keep = ((vi == label) | (vj == label)) & boundary
        # The result of shapely polygonize is always a GeometryCollection.
        # Holes are included twice: once as holes in the largest body, and once
        # more as polygons on their own. We are interested in the largest
        # polygon, which we identify through its bounding box.
        edges = grid.edge_node_connectivity[keep]
        collection = shapely.polygonize(shapely.linestrings(coordinates[edges]))
        polygon = max(collection.geoms, key=lambda x: _bbox_area(x.bounds))
        # Find the first True value in keep, use that to fetch the polygon
        # value.
        value = data_i[keep.argmax()]
        polygons.append(polygon)
        values.append(value)

    return gpd.GeoDataFrame({"values": values}, geometry=polygons)
